@@ -15,10 +15,10 @@
 #endif
 
 #include <link-grammar/api-structures.h>
-#include "corpus/corpus.h"
-#include "error.h"
+#include "link-grammar/corpus/corpus.h"
+#include "link-grammar/error.h"
 #include "jni-client.h"
-#include "utilities.h"
+#include "link-grammar/utilities.h"
 
 /* Make the compiler shut up about the deprecated functions */
 #pragma GCC diagnostic push
@@ -277,13 +277,6 @@ static void makeLinkage(per_thread_data *ptd)
 			linkage_delete(ptd->linkage);
 
 		ptd->linkage = linkage_create(ptd->cur_linkage, ptd->sent, ptd->opts);
-
-#ifdef USE_FAT_LINKAGES
-		if (!ptd->linkage) return;
-		linkage_compute_union(ptd->linkage);
-		linkage_set_current_sublinkage(ptd->linkage,
-		                        linkage_get_num_sublinkages(ptd->linkage)-1);
-#endif /* USE_FAT_LINKAGES */
 	}
 }
 
@@ -467,15 +460,13 @@ Java_org_linkgrammar_LinkGrammar_getLinkageSense(JNIEnv *env,
 {
 	per_thread_data *ptd = get_ptd(env, cls);
 	Linkage lkg = ptd->linkage;
-	Linkage_info *lifo;
 	Sense *sns;
 	const char * w = NULL;
 	jstring js;
 
 	if (!lkg) return NULL;
-	lifo = lkg->info;
 	lg_corpus_linkage_senses(lkg);
-	sns = lg_get_word_sense(lifo, i);
+	sns = lg_get_word_sense(&lkg->lifo, i);
 	while ((0 < j) && sns)
 	{
 		sns = lg_sense_next(sns);
@@ -496,13 +487,11 @@ Java_org_linkgrammar_LinkGrammar_getLinkageSenseScore(JNIEnv *env,
 {
 	per_thread_data *ptd = get_ptd(env, cls);
 	Linkage lkg = ptd->linkage;
-	Linkage_info *lifo = lkg->info;
 	Sense *sns;
 	double score = 0.0;
 
 	if (!lkg) return 0.0;
-	lifo = lkg->info;
-	sns = lg_get_word_sense(lifo, i);
+	sns = lg_get_word_sense(&lkg->lifo, i);
 	while ((0 < j) && sns)
 	{
 		sns = lg_sense_next(sns);
