@@ -1,6 +1,7 @@
 /*************************************************************************/
 /* Copyright (c) 2004                                                    */
 /* Daniel Sleator, David Temperley, and John Lafferty                    */
+/* Copyright (c) 2014 Linas Vepstas                                      */
 /* All rights reserved                                                   */
 /*                                                                       */
 /* Use of the link grammar parsing system is subject to the terms of the */
@@ -41,7 +42,7 @@
 struct Cost_Model_s
 {
 	Cost_Model_type type;
-	int (*compare_fn)(Linkage_info *, Linkage_info *);
+	int (*compare_fn)(Linkage, Linkage);
 };
 
 struct Resources_s
@@ -85,10 +86,6 @@ struct Parse_Options_s
 	bool all_short;        /* If true, there can be no connectors that are exempt */
 	bool repeatable_rand;  /* Reset rand number gen after every parse. */
 
-#ifdef USE_FAT_LINKAGES
-	bool use_fat_links;     /* Look for fat linkages */
-#endif /* USE_FAT_LINKAGES */
-
 	/* Options governing post-processing */
 	size_t twopass_length; /* min sent length for two-pass post processing */
 	Cost_Model cost_model; /* For sorting linkages in post_processing */
@@ -111,6 +108,12 @@ struct Afdict_class_struct
 	char const ** string;
 };
 
+/* Used for memory management */
+struct Exp_list_s
+{
+	Exp * exp_list;
+};
+
 struct Dictionary_s
 {
 	Dict_node *     root;
@@ -118,6 +121,7 @@ struct Dictionary_s
 	const char *    name;
 	const char *    lang;
 	const char *    version;
+	int             num_entries;
 
 	bool         use_unknown_word;
 	bool         unknown_word_defined;
@@ -147,56 +151,29 @@ struct Dictionary_s
 	bool (*lookup)(Dictionary, const char*);
 	void (*close)(Dictionary);
 
-
-	Postprocessor * postprocessor;
-	Postprocessor * constituent_pp;
-#ifdef USE_FAT_LINKAGES
-	Connector_set * andable_connector_set;  /* NULL=everything is andable */
-#endif /* USE_FAT_LINKAGES */
+	pp_knowledge  * base_knowledge;    /* Core post-processing rules */
+	pp_knowledge  * hpsg_knowledge;    /* Head-Phrase Structure rules */
 	Connector_set * unlimited_connector_set; /* NULL=everthing is unlimited */
-	String_set *    string_set;  /* Set of link names constructed during parsing */
-	int             num_entries;
+	String_set *    string_set;   /* Set of link names in the dictionary */
 	Word_file *     word_file_header;
 
 	/* exp_list links together all the Exp structs that are allocated
 	 * in reading this dictionary.  Needed for freeing the dictionary
 	 */
-	Exp *           exp_list;
+	Exp_list        exp_list;
 
 	/* Private data elements that come in play only while the
 	 * dictionary is being read, and are not otherwise used.
 	 */
 	const char    * input;
 	const char    * pin;
-	char            token[MAX_TOKEN_LENGTH];
 	bool            recursive_error;
 	bool            is_special;
 	char            already_got_it;
 	int             line_number;
+	char            token[MAX_TOKEN_LENGTH];
 };
 
-struct Label_node_s
-{
-	int          label;
-	Label_node * next;
-};
-
-#define HT_SIZE (1<<10)
-
-#ifdef USE_FAT_LINKAGES
-struct And_data_s
-{
-	int          LT_bound;
-	int          LT_size;
-	Disjunct **  label_table;
-	Label_node * hash_table[HT_SIZE];
-
-	/* keeping statistics */
-	int STAT_N_disjuncts;
-	int STAT_calls_to_equality_test;
-};
-#endif /* USE_FAT_LINKAGES */
-
 struct Link_s
 {
 	size_t lw;              /* Offset into Linkage->word NOT Sentence->word */
@@ -210,61 +187,32 @@ struct Link_s
  * N_words:
  *   The number of words in the current sentence.  Computed by
  *   separate_sentence().
- *
- * N_links:
- *   The number of links in the current linkage.  Computed by
- *   extract_linkage().
- *
- * chosen_disjuncts[]
- *   This is an array pointers to disjuncts, one for each word, that is
- *   computed by extract_links().  It represents the chosen disjuncts
- *   for the current linkage.  It is used to compute the cost of the
- *   linkage, and also by compute_chosen_words() to compute the
- *   chosen_words[].
- *
- * link_array[]
- *   This is an array of links.  These links define the current linkage.
- *    It is computed by extract_links().  It is used by analyze_linkage()
- *    to compute pp_linkage[].
  */
 struct Parse_info_struct
 {
 	unsigned int   x_table_size;
 	unsigned int   log2_x_table_size;
-	X_table_connector ** x_table;
+	X_table_connector ** x_table;  /* Hash table */
 	Parse_set *    parse_set;
 	int            N_words;
-	Disjunct **    chosen_disjuncts;
-	size_t         N_links;
-	size_t         lasz;
-	Link           *link_array;
-
-#ifdef USE_FAT_LINKAGES
-	/* Points to the image structure for each word.
-	 * NULL if not a fat word. */
-	Image_node ** image_array;
-
-	/* Array of boolean flags, one per word. Set to TRUE if this 
-	 * word has a fat down link. FALSE otherise */
-	bool *has_fat_down;
-#endif /* USE_FAT_LINKAGES */
 
 	/* thread-safe random number state */
 	unsigned int rand_state;
 };
 
+
 struct Sentence_s
 {
-	Dictionary  dict;           /* words are defined from this dictionary */
+	Dictionary  dict;           /* Words are defined from this dictionary */
 	const char *orig_sentence;  /* Copy of original sentence */
-	size_t length;              /* number of words */
-	Word  *word;                /* array of words after tokenization */
-	String_set *   string_set;  /* used for word names, not connectors */
+	size_t length;              /* Number of words */
+	Word  *word;                /* Array of words after tokenization */
+	String_set *   string_set;  /* Used for assorted strings */
 
 	/* Parse results */
-	int    num_linkages_found;  /* total number before postprocessing.  This
+	int    num_linkages_found;  /* Total number before postprocessing.  This
 	                               is returned by the count() function */
-	size_t num_linkages_alloced;/* total number of linkages allocated.
+	size_t num_linkages_alloced;/* Total number of linkages allocated.
 	                               the number post-processed might be fewer
 	                               because some are non-canonical */
 	size_t num_linkages_post_processed;
@@ -272,35 +220,20 @@ struct Sentence_s
 	                               put into the array that was alloced.
 	                               This is not the same as num alloced
 	                               because some may be non-canonical. */
-	size_t num_valid_linkages;  /* number with no pp violations */
-	size_t null_count;          /* number of null links in linkages */
-	Parse_info     parse_info;  /* set of parses for the sentence */
-	Linkage_info * link_info;   /* array of valid and invalid linkages (sorted) */
+	size_t num_valid_linkages;  /* Number with no pp violations */
+	size_t null_count;          /* Number of null links in linkages */
+	Linkage        lnkages;     /* Sorted array of valid & invalid linkages */
+	Postprocessor * postprocessor;
+	Postprocessor * constituent_pp;
+
+	/* parse_info not used by SAT solver */
+	Parse_info     parse_info;  /* Set of parses for the sentence */
 
 	/* Tokenizer internal/private state */
 	bool   * post_quote;        /* Array, one entry per word, true if quote */
 	int    t_start;             /* start word of the current token sequence */
 	int    t_count;             /* word count in the current token sequence */
 
-#ifdef USE_FAT_LINKAGES
-	/* parser internal/private state */
-	/* Obsolete fat-linkage state */
-	analyze_context_t * analyze_ctxt; /* private state  used for analyzing */
-	count_context_t * count_ctxt; /* private state info used for counting */
-
-	bool   null_links;          /* null links allowed */
-	int    num_thin_linkages;   /* valid linkages which are not fat */
-	char * is_conjunction;      /* Array of flags, one per word; set to
-	                               TRUE if conjunction, as defined by dictionary */
-	char** deletable;           /* deletable regions in a sentence with conjunction */
-	char** dptr;                /* private pointer for mem management only */
-	char** effective_dist;
-	And_data       and_data;    /* used to keep track of fat disjuncts */
-#endif /* USE_FAT_LINKAGES */
-
-	/* Post-processor private/internal state */
-	bool  q_pruned_rules;       /* don't prune rules more than once in p.p. */
-
 	/* thread-safe random number state */
 	unsigned int rand_state;
 
@@ -339,11 +272,13 @@ struct DTreeLeaf_s
 
 struct PP_data_s
 {
+	List_o_links ** word_links;
+	size_t wowlen;
 	size_t N_domains;
-	List_o_links * word_links[MAX_SENTENCE];
+	Domain * domain_array;          /* The domains, sorted by size */
+	size_t domlen;                  /* Allocated size of domain_array */
+	size_t length;                  /* Length of current linkage */
 	List_o_links * links_to_ignore;
-	Domain domain_array[MAX_LINKS]; /* the domains, sorted by size */
-	size_t length;                  /* length of current sentence */
 };
 
 struct PP_info_s
@@ -354,68 +289,84 @@ struct PP_info_s
 
 struct Postprocessor_s
 {
-	pp_knowledge *knowledge;             /* internal rep'n of the actual rules */
+	pp_knowledge  * knowledge;           /* Internal rep'n of the actual rules */
 	int n_global_rules_firing;           /* this & the next are diagnostic     */
 	int n_local_rules_firing;
 	pp_linkset *set_of_links_of_sentence;     /* seen in *any* linkage of sent */
 	pp_linkset *set_of_links_in_an_active_rule;/*used in *some* linkage of sent*/
 	int *relevant_contains_one_rules;        /* -1-terminated list of indices  */
 	int *relevant_contains_none_rules;
+	bool q_pruned_rules;       /* don't prune rules more than once in p.p. */
 
-	/* the following maintain state during a call to post_process() */
-	String_set *sentence_link_name_set;        /* link names seen for sentence */
-	bool visited[MAX_SENTENCE];                 /* for the depth-first search */
+	/* The following maintain state during a call to post_process() */
+	String_set *string_set;      /* Link names seen for sentence */
+	bool *visited;               /* For the depth-first search */
+	size_t vlength;              /* Length of visited array */
 	PP_node *pp_node;
 	PP_data pp_data;
 };
 
-
 /*********************************************************
  *
  * Linkages
  *
  **********************************************************/
 
-/* XXX FIXME When fat links are removed, there will only ever be
- * just one sublinkage, and so this struct can be merged into 
- * Linkage_s below, and removed from the API. Well, sort-of. The
- * analyze_thin_linkage() routine still wants to malloc a temp
- * Sublinkage struct, for post-processing.  So we should wait until
- * post-processing is removed....
+/**
+ * This is for building the graphs of links in post-processing.
  */
-struct Sublinkage_s
+struct Linkage_info_struct
 {
-	size_t    num_links;          /* Number of links in array */
-	Link **   link;               /* Array of links */
-	PP_info * pp_info;            /* PP info for each link */
-	const char * violation;       /* Name of violation, if any */
-	PP_data   pp_data;
+	int index;            /* Index into the parse_set */
+	bool discarded;
+	short N_violations;
+	short unused_word_cost;
+	short link_cost;
+
+	double disjunct_cost;
+	double corpus_cost;
+	size_t nwords;
+	const char *pp_violation_msg;
+	char **disjunct_list_str;
+#ifdef USE_CORPUS
+	Sense **sense_list;
+#endif
 };
 
+/**
+ * num_links:
+ *   The number of links in the current linkage.  Computed by
+ *   extract_linkage().
+ *
+ * chosen_disjuncts[]
+ *   This is an array pointers to disjuncts, one for each word, that is
+ *   computed by extract_links().  It represents the chosen disjuncts
+ *   for the current linkage.  It is used to compute the cost of the
+ *   linkage, and also by compute_chosen_words() to compute the
+ *   chosen_words[].
+ *
+ * link_array[]
+ *   This is an array of links.  These links define the current linkage.
+ *   It is computed by extract_links().  It is used by analyze_linkage().
+ */
 struct Linkage_s
 {
-	size_t          num_words;  /* number of (tokenized) words */
-	const char *  * word;       /* array of word spellings */
-	Linkage_info*   info;       /* index and cost information */
-#ifdef USE_FAT_LINKAGES
-	Sublinkage *    sublinkage; /* A parse with conjunctions will have several */
-#else
-	Sublinkage      sublinkage; /* Just one */
-#endif /* USE_FAT_LINKAGES */
+	size_t          num_words;    /* Number of (tokenized) words */
+	const char *  * word;         /* Array of word spellings */
 
-	Sentence        sent;
-	Parse_Options   opts;
+	size_t          num_links;    /* Number of links in array */
+	Link *          link_array;   /* Array of links */
+	size_t          lasz;         /* Alloc'ed length of link_array */
 
-#ifdef USE_FAT_LINKAGES
-	int             num_sublinkages; /* One for thin linkages, bigger for fat */
-	int             current;    /* Allows user to select particular sublinkage */
-	bool            unionized;  /* if TRUE, union of links has been computed */
-	DIS_node      * dis_con_tree; /* Disjunction-conjunction tree */
-#endif /* USE_FAT_LINKAGES */
+	Disjunct **     chosen_disjuncts; /* Disjuncts used, one per word */
+	size_t          cdsz;         /* Alloc'ed length of chosen_disjuncts */
 
+	Linkage_info    lifo;         /* Parse_set index and cost information */
+	PP_info *       pp_info;      /* PP info, one for each link */
+	const char *    pp_violation; /* Name of violation, if any */
+	PP_data         hpsg_pp_data; /* Used in constituent code */
 };
 
 
-
 #endif
 
