@@ -25,12 +25,14 @@
 #include "externs.h"
 #include "extract-links.h"
 #include "fast-match.h"
+#include "linkage.h"
 #include "post-process.h"
 #include "preparation.h"
 #include "print.h"
 #include "prune.h"
 #include "regex-morph.h"
 #include "resources.h"
+#include "score.h"
 #include "sat-solver/sat-encoder.h"
 #include "spellcheck.h"
 #include "string-set.h"
@@ -39,10 +41,6 @@
 #include "utilities.h"
 #include "word-utils.h"
 
-#ifdef USE_FAT_LINKAGES
-#include "and.h"
-#endif /* USE_FAT_LINKAGES */
-
 /***************************************************************
 *
 * Routines for setting Parse_Options
@@ -53,8 +51,11 @@
  * For sorting the linkages in postprocessing
  */
 
-static int VDAL_compare_parse(Linkage_info * p1, Linkage_info * p2)
+static int VDAL_compare_parse(Linkage l1, Linkage l2)
 {
+	Linkage_info * p1 = &l1->lifo;
+	Linkage_info * p2 = &l2->lifo;
+
 	/* Move the discarded entries to the end of the list */
 	if (p1->discarded || p2->discarded) return (p1->discarded - p2->discarded);
 
@@ -64,26 +65,19 @@ static int VDAL_compare_parse(Linkage_info * p1, Linkage_info * p2)
 	else if (p1->unused_word_cost != p2->unused_word_cost) {
 		return (p1->unused_word_cost - p2->unused_word_cost);
 	}
-#ifdef USE_FAT_LINKAGES
-	else if (p1->fat != p2->fat) {
-		return (p1->fat - p2->fat);
-	}
-#endif /* USE_FAT_LINKAGES */
 	else if (p1->disjunct_cost > p2->disjunct_cost) return 1;
 	else if (p1->disjunct_cost < p2->disjunct_cost) return -1;
-#ifdef USE_FAT_LINKAGES
-	else if (p1->and_cost != p2->and_cost) {
-		return (p1->and_cost - p2->and_cost);
-	}
-#endif /* USE_FAT_LINKAGES */
 	else {
 		return (p1->link_cost - p2->link_cost);
 	}
 }
 
 #ifdef USE_CORPUS
-static int CORP_compare_parse(Linkage_info * p1, Linkage_info * p2)
+static int CORP_compare_parse(Linkage l1, Linkage l2)
 {
+	Linkage_info * p1 = &l1->lifo;
+	Linkage_info * p2 = &l2->lifo;
+
 	double diff = p1->corpus_cost - p2->corpus_cost;
 
 	/* Move the discarded entries to the end of the list */
@@ -140,9 +134,6 @@ Parse_Options parse_options_create(void)
 	po->twopass_length = 30;
 	po->repeatable_rand = true;
 	po->resources = resources_create();
-#ifdef USE_FAT_LINKAGES
-	po->use_fat_links = false;
-#endif /* USE_FAT_LINKAGES */
 	po->use_cluster_disjuncts = false;
 	po->display_morphology = false;
 
@@ -287,15 +278,6 @@ bool parse_options_get_use_viterbi(Parse_Options opts) {
 	return opts->use_viterbi;
 }
 
-#ifdef USE_FAT_LINKAGES
-void parse_options_set_use_fat_links(Parse_Options opts, int dummy) {
-	opts->use_fat_links = dummy;
-}
-int parse_options_get_use_fat_links(Parse_Options opts) {
-	return opts->use_fat_links;
-}
-#endif /* USE_FAT_LINKAGES */
-
 void parse_options_set_linkage_limit(Parse_Options opts, int dummy)
 {
 	opts->linkage_limit = dummy;
@@ -422,65 +404,42 @@ void parse_options_reset_resources(Parse_Options opts) {
 *
 ****************************************************************/
 
-static Linkage_info * linkage_info_new(int num_to_alloc)
+static Linkage linkage_array_new(int num_to_alloc)
 {
-	Linkage_info *link_info;
-	link_info = (Linkage_info *) xalloc(num_to_alloc * sizeof(Linkage_info));
-	memset(link_info, 0, num_to_alloc * sizeof(Linkage_info));
-	return link_info;
+	Linkage lkgs = (Linkage) exalloc(num_to_alloc * sizeof(struct Linkage_s));
+	memset(lkgs, 0, num_to_alloc * sizeof(struct Linkage_s));
+	return lkgs;
 }
 
-static void linkage_info_delete(Linkage_info *link_info, int sz)
+static void free_linkages(Sentence sent)
 {
-	int i,j;
+	size_t in;
+	Linkage lkgs = sent->lnkages;
+	if (!lkgs) return;
 
-	for (i=0; i<sz; i++)
+	for (in=0; in<sent->num_linkages_alloced; in++)
 	{
-		Linkage_info *lifo = &link_info[i];
-		int nwords = lifo->nwords;
-		for (j=0; j<nwords; j++)
+		size_t j;
+		Linkage linkage = &lkgs[in];
+		exfree((void *) linkage->word, sizeof(const char *) * linkage->num_words);
+		exfree(linkage->chosen_disjuncts, linkage->num_words * sizeof(Disjunct *));
+		free(linkage->link_array);
+
+		// XXX why isn't this in a string set ??
+		for (j=0; j<linkage->lifo.nwords; j++)
 		{
-			if (lifo->disjunct_list_str[j])
-				free(lifo->disjunct_list_str[j]);
+			if (linkage->lifo.disjunct_list_str[j])
+				free(linkage->lifo.disjunct_list_str[j]);
 		}
-		free(lifo->disjunct_list_str);
+		free(linkage->lifo.disjunct_list_str);
 #ifdef USE_CORPUS
-		lg_sense_delete(lifo);
+		lg_sense_delete(&linkage->lifo);
 #endif
-	}
-	xfree(link_info, sz * sizeof(Linkage_info));
-}
 
-#ifdef USE_FAT_LINKAGES
-static void free_andlists(Sentence sent)
-{
-	size_t L;
-	Andlist * andlist, * next;
-	for (L=0; L<sent->num_linkages_post_processed; L++) {
-		/* printf("%d ", sent->link_info[L].canonical);  */
-		/* if (sent->link_info[L].canonical==0) continue; */
-		andlist = sent->link_info[L].andlist;
-		while(1) {
-			if(andlist == NULL) break;
-			next = andlist->next;
-			xfree((char *) andlist, sizeof(Andlist));
-			andlist = next;
-		}
+		linkage_free_pp_info(linkage);
 	}
-	/* printf("\n"); */
-}
-#endif /* USE_FAT_LINKAGES */
 
-static void free_post_processing(Sentence sent)
-{
-	if (sent->link_info != NULL) {
-		/* postprocessing must have been done */
-#ifdef USE_FAT_LINKAGES
-		free_andlists(sent);
-#endif /* USE_FAT_LINKAGES */
-		linkage_info_delete(sent->link_info, sent->num_linkages_alloced);
-		sent->link_info = NULL;
-	}
+	exfree(lkgs, sent->num_linkages_alloced * sizeof(struct Linkage_s));
 }
 
 static void select_linkages(Sentence sent, fast_matcher_t* mchxt,
@@ -489,12 +448,8 @@ static void select_linkages(Sentence sent, fast_matcher_t* mchxt,
 {
 	size_t in;
 	size_t N_linkages_found, N_linkages_alloced;
-	bool overflowed;
-	Linkage_info *link_info;
 
-	free_post_processing(sent);
-
-	overflowed = build_parse_set(sent, mchxt, ctxt, sent->null_count, opts);
+	bool overflowed = build_parse_set(sent, mchxt, ctxt, sent->null_count, opts);
 	print_time(opts, "Built parse set");
 
 	if (overflowed && (1 < opts->verbosity))
@@ -512,10 +467,7 @@ static void select_linkages(Sentence sent, fast_matcher_t* mchxt,
 		sent->num_linkages_alloced = 0;
 		sent->num_linkages_post_processed = 0;
 		sent->num_valid_linkages = 0;
-#ifdef USE_FAT_LINKAGES
-		sent->num_thin_linkages = 0;
-#endif /* USE_FAT_LINKAGES */
-		sent->link_info = NULL;
+		sent->lnkages = NULL;
 		return;
 	}
 
@@ -536,18 +488,28 @@ static void select_linkages(Sentence sent, fast_matcher_t* mchxt,
 		N_linkages_alloced = N_linkages_found;
 	}
 
-	link_info = linkage_info_new(N_linkages_alloced);
+	/* Now actually malloc the array in which we will process linkages. */
+	sent->lnkages = linkage_array_new(N_linkages_alloced);
 
 	/* Generate an array of linkage indices to examine */
 	if (overflowed)
 	{
+		/* The negative index means that a random subset of links
+		 * will be picked later on, in extract_links(). */
 		for (in=0; in < N_linkages_alloced; in++)
 		{
-			link_info[in].index = -(in+1);
+			sent->lnkages[in].lifo.index = -(in+1);
 		}
 	}
+	else if (N_linkages_found == N_linkages_alloced)
+	{
+		for (in=0; in<N_linkages_alloced; in++)
+			sent->lnkages[in].lifo.index = in;
+	}
 	else
 	{
+		/* There are more linkages found than we can handle */
+		/* Pick a (quasi-)uniformly distributed random subset. */
 		if (opts->repeatable_rand)
 			sent->rand_state = N_linkages_found + sent->length;
 
@@ -559,49 +521,70 @@ static void select_linkages(Sentence sent, fast_matcher_t* mchxt,
 			frac /= (double) N_linkages_alloced;
 			block_bottom = (int) (((double) in) * frac);
 			block_top = (int) (((double) (in+1)) * frac);
-			link_info[in].index = block_bottom +
+			sent->lnkages[in].lifo.index = block_bottom +
 				(rand_r(&sent->rand_state) % (block_top-block_bottom));
 		}
 	}
 
-	sent->link_info = link_info;
 	sent->num_linkages_alloced = N_linkages_alloced;
-	/* Later we subtract the number of invalid linkages */
+	/* Later, we subtract the number of invalid linkages */
 	sent->num_valid_linkages = N_linkages_alloced;
-
-	/* Allow use of both old and new pp order, for tests.
-	 * FIXME In release code this assignment can be removed if sane_morphism()
-	 * is modified to loop over num_linkages_alloced. */
-	sent->num_linkages_post_processed = N_linkages_alloced;
 }
 
+/* Partial, but not full initialization of the linakge struct ... */
+void partial_init_linkage(Linkage lkg, unsigned int N_words)
+{
+	lkg->num_links = 0;
+	lkg->lasz = 2 * N_words;
+	lkg->link_array = (Link *) malloc(lkg->lasz * sizeof(Link));
+	memset(lkg->link_array, 0, lkg->lasz * sizeof(Link));
+
+	lkg->num_words = N_words;
+	lkg->cdsz =  N_words;
+	lkg->chosen_disjuncts = (Disjunct **) exalloc(lkg->cdsz * sizeof(Disjunct *));
+	memset(lkg->chosen_disjuncts, 0, N_words * sizeof(Disjunct *));
+
+	lkg->pp_info = NULL;
+	lkg->pp_violation = NULL;
+}
+
+void check_link_size(Linkage lkg)
+{
+	if (lkg->lasz <= lkg->num_links)
+	{
+		lkg->lasz = 2 * lkg->lasz + 10;
+		lkg->link_array = realloc(lkg->link_array, lkg->lasz * sizeof(Link));
+	}
+}
+
+/** The extract_links() call sets the chosen_disjuncts array */
+static void compute_chosen_disjuncts(Sentence sent)
+{
+	size_t in;
+	size_t N_linkages_alloced = sent->num_linkages_alloced;
+	Parse_info pi = sent->parse_info;
+
+	for (in=0; in < N_linkages_alloced; in++)
+	{
+		Linkage lkg = &sent->lnkages[in];
+		Linkage_info *lifo = &lkg->lifo;
+		if (lifo->discarded || lifo->N_violations) continue;
+
+		partial_init_linkage(lkg, pi->N_words);
+		extract_links(lkg, pi);
+	}
+}
+
+/** This does basic post-processing for all linkages.
+ */
 static void post_process_linkages(Sentence sent, Parse_Options opts)
 {
-	Linkage_info *lifo;
 	size_t in;
 	size_t N_linkages_post_processed = 0;
 	size_t N_valid_linkages = sent->num_valid_linkages;
 	size_t N_linkages_alloced = sent->num_linkages_alloced;
-#ifdef USE_FAT_LINKAGES
-	bool overflowed = (sent->link_info[0].index < 0);
-	size_t N_linkages_found = sent->num_linkages_found;
-	size_t N_thin_linkages;
-	bool only_canonical_allowed;
-	bool canonical;
-#endif /* USE_FAT_LINKAGES */
 
-#ifdef USE_FAT_LINKAGES
-	/* When we're processing only a small subset of the linkages,
-	 * don't worry about restricting the set we consider to be
-	 * canonical ones.  In the extreme case where we are only
-	 * generating 1 in a million linkages, it's very unlikely
-	 * that we'll hit two symmetric variants of the same linkage
-	 * anyway.
-	 */
-	only_canonical_allowed = !(overflowed || (N_linkages_found > 2*opts->linkage_limit));
-#endif /* USE_FAT_LINKAGES */
-
-	/* (optional) first pass: just visit the linkages */
+	/* (optional) First pass: just visit the linkages */
 	/* The purpose of these two passes is to make the post-processing
 	 * more efficient.  Because (hopefully) by the time you do the
 	 * real work in the 2nd pass you've pruned the relevant rule set
@@ -611,32 +594,28 @@ static void post_process_linkages(Sentence sent, Parse_Options opts)
 	{
 		for (in=0; in < N_linkages_alloced; in++)
 		{
-			lifo = &sent->link_info[in];
+			Linkage lkg = &sent->lnkages[in];
+			Linkage_info *lifo = &lkg->lifo;
 			if (lifo->discarded || lifo->N_violations) continue;
-			extract_links(lifo->index, sent->parse_info);
-#ifdef USE_FAT_LINKAGES
-			if (set_has_fat_down(sent))
-			{
-				if (only_canonical_allowed && !is_canonical_linkage(sent)) continue;
-				analyze_fat_linkage(sent, opts, PP_FIRST_PASS);
-			}
-			else
-#endif /* USE_FAT_LINKAGES */
-			{
-				analyze_thin_linkage(sent, opts, PP_FIRST_PASS);
-			}
+
+			compute_link_names(lkg, sent->string_set);
+			post_process_scan_linkage(sent->postprocessor, opts, lkg);
+
 			if ((9 == in%10) && resources_exhausted(opts->resources)) break;
 		}
 	}
 
-	/* second pass: actually perform post-processing */
-#ifdef USE_FAT_LINKAGES
-	N_thin_linkages = 0;
-#endif /* USE_FAT_LINKAGES */
+	/* Second pass: actually perform post-processing */
 	for (in=0; in < N_linkages_alloced; in++)
 	{
-		lifo = &sent->link_info[in];
+		PP_node *ppn;
+		Linkage lkg = &sent->lnkages[in];
+		Linkage_info *lifo = &lkg->lifo;
+
 		if (lifo->discarded) continue;
+
+		/* N_violations could be non-zero if the linkage failed the
+		 * morphology check */
 		if (lifo->N_violations)
 		{
 			/* Arrange for displaying "Invalid morphism construction" sentences
@@ -644,75 +623,27 @@ static void post_process_linkages(Sentence sent, Parse_Options opts)
 			N_linkages_post_processed++;
 			continue;
 		}
-		extract_links(lifo->index, sent->parse_info);
-#ifdef USE_FAT_LINKAGES
-		lifo->fat = false;
-		lifo->canonical = true;
-		if (set_has_fat_down(sent))
-		{
-			int index = lifo->index;
-			canonical = is_canonical_linkage(sent);
-			if (only_canonical_allowed && !canonical)
-			{
-				lifo->discarded = true;
-				continue;
-			}
-			*lifo = analyze_fat_linkage(sent, opts, PP_SECOND_PASS);
-			lifo->index = index;
-			lifo->fat = true;
-			lifo->canonical = canonical;
-		}
-		else
-#endif /* USE_FAT_LINKAGES */
-		{
-			int index = lifo->index;
-			*lifo = analyze_thin_linkage(sent, opts, PP_SECOND_PASS);
-			lifo->index = index;
-		}
-		if (0 == lifo->N_violations)
-		{
-#ifdef USE_FAT_LINKAGES
-			if (false == lifo->fat)
-				N_thin_linkages++;
-#endif /* USE_FAT_LINKAGES */
-		}
-		else
+		compute_link_names(lkg, sent->string_set);
+		ppn = do_post_process(sent->postprocessor, opts, lkg);
+	   post_process_free_data(&sent->postprocessor->pp_data);
+
+		if (NULL != ppn->violation)
 		{
 			N_valid_linkages--;
+			lifo->N_violations++;
+
+			/* Set the message, only if not set (e.g. by sane_morphism) */
+			if (NULL == lifo->pp_violation_msg)
+				lifo->pp_violation_msg = ppn->violation;
 		}
-		lg_corpus_score(sent, lifo);
 		N_linkages_post_processed++;
+
+		linkage_score(lkg, opts);
 		if ((9 == in%10) && resources_exhausted(opts->resources)) break;
 	}
 
 	print_time(opts, "Postprocessed all linkages");
 
-#ifdef USE_FAT_LINKAGES
-	if (!resources_exhausted(opts->resources))
-	{
-		if ((N_linkages_post_processed == 0) &&
-		    (N_linkages_found > 0) &&
-		    (N_linkages_found <= opts->linkage_limit))
-		{
-			/* With the current parser, the following sentence will elicit
-			 * this error:
-			 *
-			 * Well, say, Joe, you can be Friar Tuck or Much the miller's
-			 * son, and lam me with a quarter-staff; or I'll be the Sheriff
-			 * of Nottingham and you be Robin Hood a little while and kill
-			 * me.
-			 */
-			err_ctxt ec;
-			ec.sent = sent;
-			err_msg(&ec, Error, "Error: None of the linkages is canonical\n"
-			          "\tN_linkages_post_processed=%zu "
-			          "N_linkages_found=%zu\n",
-			          N_linkages_post_processed,
-			          N_linkages_found);
-		}
-	}
-#endif /* USE_FAT_LINKAGES */
-
 	if (opts->verbosity > 1)
 	{
 		err_ctxt ec;
@@ -724,15 +655,12 @@ static void post_process_linkages(Sentence sent, Parse_Options opts)
 	sent->num_linkages_alloced = N_linkages_alloced;
 	sent->num_linkages_post_processed = N_linkages_post_processed;
 	sent->num_valid_linkages = N_valid_linkages;
-#ifdef USE_FAT_LINKAGES
-	sent->num_thin_linkages = N_thin_linkages;
-#endif /* USE_FAT_LINKAGES */
 }
 
 static void sort_linkages(Sentence sent, Parse_Options opts)
 {
-	qsort((void *)sent->link_info, sent->num_linkages_alloced,
-	      sizeof(Linkage_info),
+	qsort((void *)sent->lnkages, sent->num_linkages_alloced,
+	      sizeof(struct Linkage_s),
 	      (int (*)(const void *, const void *))opts->cost_model.compare_fn);
 
 #if 0
@@ -742,7 +670,7 @@ static void sort_linkages(Sentence sent, Parse_Options opts)
 		size_t N_linkages_post_processed = 0;
 		for (in=0; in < sent->num_linkages_alloced; in++)
 		{
-			Linkage_info *lifo = &sent->link_info[in];
+			Linkage_info *lifo = &sent->lnkages[in].lifo;
 			if (lifo->discarded) break;
 			N_linkages_post_processed++;
 		}
@@ -755,223 +683,6 @@ static void sort_linkages(Sentence sent, Parse_Options opts)
 	print_time(opts, "Sorted all linkages");
 }
 
-static void old_post_process_linkages(Sentence sent, fast_matcher_t* mchxt,
-                                  count_context_t* ctxt,
-                                  Parse_Options opts)
-{
-	int *indices;
-	size_t in, block_bottom, block_top;
-	size_t N_linkages_found, N_linkages_alloced;
-	size_t N_linkages_post_processed, N_valid_linkages;
-	bool overflowed;
-	Linkage_info *link_info;
-#ifdef USE_FAT_LINKAGES
-	size_t N_thin_linkages;
-	bool only_canonical_allowed;
-	bool canonical;
-#endif /* USE_FAT_LINKAGES */
-
-	free_post_processing(sent);
-
-	overflowed = build_parse_set(sent, mchxt, ctxt, sent->null_count, opts);
-	print_time(opts, "Built parse set");
-
-	if (overflowed && (1 < opts->verbosity))
-	{
-		err_ctxt ec;
-		ec.sent = sent;
-		err_msg(&ec, Warn, "Warning: Count overflow.\n"
-		  "Considering a random subset of %zu of an unknown and large number of linkages\n",
-			opts->linkage_limit);
-	}
-	N_linkages_found = sent->num_linkages_found;
-
-	if (sent->num_linkages_found == 0)
-	{
-		sent->num_linkages_alloced = 0;
-		sent->num_linkages_post_processed = 0;
-		sent->num_valid_linkages = 0;
-#ifdef USE_FAT_LINKAGES
-		sent->num_thin_linkages = 0;
-#endif /* USE_FAT_LINKAGES */
-		sent->link_info = NULL;
-		return;
-	}
-
-	if (N_linkages_found > opts->linkage_limit)
-	{
-		N_linkages_alloced = opts->linkage_limit;
-		if (opts->verbosity > 1)
-		{
-			err_ctxt ec;
-			ec.sent = sent;
-			err_msg(&ec, Warn,
-			    "Warning: Considering a random subset of %zu of %zu linkages\n",
-			    N_linkages_alloced, N_linkages_found);
-		}
-	}
-	else
-	{
-		N_linkages_alloced = N_linkages_found;
-	}
-
-	link_info = linkage_info_new(N_linkages_alloced);
-	N_valid_linkages = 0;
-
-	/* Generate an array of linkage indices to examine */
-	indices = (int *) xalloc(N_linkages_alloced * sizeof(int));
-	if (overflowed)
-	{
-		for (in=0; in < N_linkages_alloced; in++)
-		{
-			indices[in] = -(in+1);
-		}
-	}
-	else
-	{
-		if (opts->repeatable_rand)
-			sent->rand_state = N_linkages_found + sent->length;
-
-		for (in=0; in<N_linkages_alloced; in++)
-		{
-			double frac = (double) N_linkages_found;
-			frac /= (double) N_linkages_alloced;
-			block_bottom = (int) (((double) in) * frac);
-			block_top = (int) (((double) (in+1)) * frac);
-			indices[in] = block_bottom +
-				(rand_r(&sent->rand_state) % (block_top-block_bottom));
-		}
-	}
-
-#ifdef USE_FAT_LINKAGES
-	/* When we're processing only a small subset of the linkages,
-	 * don't worry about restricting the set we consider to be
-	 * canonical ones.  In the extreme case where we are only
-	 * generating 1 in a million linkages, it's very unlikely
-	 * that we'll hit two symmetric variants of the same linkage
-	 * anyway.
-	 */
-	only_canonical_allowed = !(overflowed || (N_linkages_found > 2*opts->linkage_limit));
-#endif /* USE_FAT_LINKAGES */
-
-	/* (optional) first pass: just visit the linkages */
-	/* The purpose of these two passes is to make the post-processing
-	 * more efficient.  Because (hopefully) by the time you do the
-	 * real work in the 2nd pass you've pruned the relevant rule set
-	 * in the first pass.
-	 */
-	if (sent->length >= opts->twopass_length)
-	{
-		for (in=0; in < N_linkages_alloced; in++)
-		{
-			extract_links(indices[in], sent->parse_info);
-#ifdef USE_FAT_LINKAGES
-			if (set_has_fat_down(sent))
-			{
-				if (only_canonical_allowed && !is_canonical_linkage(sent)) continue;
-				analyze_fat_linkage(sent, opts, PP_FIRST_PASS);
-			}
-			else
-#endif /* USE_FAT_LINKAGES */
-			{
-				analyze_thin_linkage(sent, opts, PP_FIRST_PASS);
-			}
-			if ((9 == in%10) && resources_exhausted(opts->resources)) break;
-		}
-	}
-
-	/* second pass: actually perform post-processing */
-	N_linkages_post_processed = 0;
-#ifdef USE_FAT_LINKAGES
-	N_thin_linkages = 0;
-#endif /* USE_FAT_LINKAGES */
-	for (in=0; in < N_linkages_alloced; in++)
-	{
-		Linkage_info *lifo = &link_info[N_linkages_post_processed];
-		extract_links(indices[in], sent->parse_info);
-#ifdef USE_FAT_LINKAGES
-		lifo->fat = false;
-		lifo->canonical = true;
-		if (set_has_fat_down(sent))
-		{
-			canonical = is_canonical_linkage(sent);
-			if (only_canonical_allowed && !canonical) continue;
-			*lifo = analyze_fat_linkage(sent, opts, PP_SECOND_PASS);
-			lifo->index = indices[in];
-			lifo->fat = true;
-			lifo->canonical = canonical;
-		}
-		else
-#endif /* USE_FAT_LINKAGES */
-		{
-			*lifo = analyze_thin_linkage(sent, opts, PP_SECOND_PASS);
-			lifo->index = indices[in];
-		}
-		if (0 == lifo->N_violations)
-		{
-			N_valid_linkages++;
-#ifdef USE_FAT_LINKAGES
-			if (false == lifo->fat)
-				N_thin_linkages++;
-#endif /* USE_FAT_LINKAGES */
-		}
-		lg_corpus_score(sent, lifo);
-		N_linkages_post_processed++;
-		if ((9 == in%10) && resources_exhausted(opts->resources)) break;
-	}
-
-	print_time(opts, "Postprocessed all linkages");
-	qsort((void *)link_info, N_linkages_post_processed, sizeof(Linkage_info),
-		  (int (*)(const void *, const void *)) opts->cost_model.compare_fn);
-
-#ifdef USE_FAT_LINKAGES
-	if (!resources_exhausted(opts->resources))
-	{
-		/* This condition can ony happen w/ fat links */
-		if ((N_linkages_post_processed == 0) &&
-		    (N_linkages_found > 0) &&
-		    (N_linkages_found < opts->linkage_limit))
-		{
-			/* With the current parser, the following sentence will elicit
-			 * this error:
-			 *
-			 * Well, say, Joe, you can be Friar Tuck or Much the miller's
-			 * son, and lam me with a quarter-staff; or I'll be the Sheriff
-			 * of Nottingham and you be Robin Hood a little while and kill
-			 * me.
-			 */
-			err_ctxt ec;
-			ec.sent = sent;
-			err_msg(&ec, Error, "Error: None of the linkages is canonical\n"
-			          "\tN_linkages_post_processed=%zu "
-			          "N_linkages_found=%zu\n",
-			          N_linkages_post_processed,
-			          N_linkages_found);
-		}
-	}
-#endif /* USE_FAT_LINKAGES */
-
-	if (opts->verbosity > 1)
-	{
-		err_ctxt ec;
-		ec.sent = sent;
-		err_msg(&ec, Info, "Info: %zu of %zu linkages with no P.P. violations\n",
-				N_valid_linkages, N_linkages_post_processed);
-	}
-
-	print_time(opts, "Sorted all linkages");
-
-	sent->num_linkages_alloced = N_linkages_alloced;
-	sent->num_linkages_post_processed = N_linkages_post_processed;
-	sent->num_valid_linkages = N_valid_linkages;
-#ifdef USE_FAT_LINKAGES
-	sent->num_thin_linkages = N_thin_linkages;
-#endif /* USE_FAT_LINKAGES */
-	sent->link_info = link_info;
-
-	xfree(indices, N_linkages_alloced * sizeof(int));
-}
-
 /***************************************************************
 *
 * Routines for creating and destroying processing Sentences
@@ -995,19 +706,14 @@ Sentence sentence_create(const char *input_string, Dictionary dict)
 	sent->num_linkages_alloced = 0;
 	sent->num_linkages_post_processed = 0;
 	sent->num_valid_linkages = 0;
-	sent->link_info = NULL;
+	sent->lnkages = NULL;
 	sent->null_count = 0;
 	sent->parse_info = NULL;
 	sent->string_set = string_set_create();
 	sent->rand_state = global_rand_state;
 
-	sent->q_pruned_rules = false;
-#ifdef USE_FAT_LINKAGES
-	sent->deletable = NULL;
-	sent->is_conjunction = NULL;
-	sent->dptr = NULL;
-	sent->effective_dist = NULL;
-#endif /* USE_FAT_LINKAGES */
+	sent->postprocessor = post_process_new(dict->base_knowledge);
+	sent->constituent_pp = post_process_new(dict->hpsg_knowledge);
 
 #ifdef USE_SAT_SOLVER
 	sent->hook = NULL;
@@ -1022,35 +728,6 @@ Sentence sentence_create(const char *input_string, Dictionary dict)
 	return sent;
 }
 
-#ifdef USE_FAT_LINKAGES
-/* XXX Extreme hack alert -- English-language words are used
- * completely naked in the C source code!!! FIXME !!!!
- * That's OK, this code is deprecated/obsolete, and will be
- * removed shortly.  We've added the Russian for now, but
- * really, the Russian dictionary needs to be fixed to support
- * conjunctions properly.
- */
-static void set_is_conjunction(Sentence sent)
-{
-	size_t w;
-	const char * s;
-	for (w=0; w<sent->length; w++) {
-		s = sent->word[w].alternatives[0];
-		sent->is_conjunction[w] =
-			/* English, obviously ... */
-			(strcmp(s, "and") == 0) ||
-			(strcmp(s, "or" ) == 0) ||
-			(strcmp(s, "but") == 0) ||
-			(strcmp(s, "nor") == 0) ||
-			/* Russian */
-			(strcmp(s, "и") == 0) ||
-			(strcmp(s, "или") == 0) ||
-			(strcmp(s, "но") == 0) ||
-			(strcmp(s, "ни") == 0);
-	}
-}
-#endif /* USE_FAT_LINKAGES */
-
 int sentence_split(Sentence sent, Parse_Options opts)
 {
 	Dictionary dict = sent->dict;
@@ -1059,23 +736,12 @@ int sentence_split(Sentence sent, Parse_Options opts)
 	 * routines depend on sent-length, which might change in different
 	 * parse-opts settings.
 	 */
-#ifdef USE_FAT_LINKAGES
-	free_deletable(sent);
-#endif /* USE_FAT_LINKAGES */
-
 	/* Tokenize */
 	if (!separate_sentence(sent, opts))
 	{
 		return -1;
 	}
 
-	sent->q_pruned_rules = false; /* for post processing */
-#ifdef USE_FAT_LINKAGES
-	sent->is_conjunction = (char *) xalloc(sizeof(char)*sent->length);
-	set_is_conjunction(sent);
-	initialize_conjunction_tables(sent);
-#endif /* USE_FAT_LINKAGES */
-
 	/* If unknown_word is not defined, then no special processing
 	 * will be done for e.g. capitalized words. */
 	if (!(dict->unknown_word_defined && dict->use_unknown_word))
@@ -1111,21 +777,12 @@ void sentence_delete(Sentence sent)
 {
 	if (!sent) return;
 	sat_sentence_delete(sent);
-#ifdef USE_FAT_LINKAGES
-	free_AND_tables(sent);
-#endif /* USE_FAT_LINKAGES */
 	free_sentence_words(sent);
 	string_set_delete(sent->string_set);
-	if (sent->parse_info) free_parse_info(sent->parse_info);
-	free_post_processing(sent);
-	post_process_close_sentence(sent->dict->postprocessor);
-
-#ifdef USE_FAT_LINKAGES
-	free_deletable(sent);
-	free_effective_dist(sent);
-	free_analyze(sent);
-	if (sent->is_conjunction) xfree(sent->is_conjunction, sizeof(char)*sent->length);
-#endif /* USE_FAT_LINKAGES */
+	free_parse_info(sent->parse_info);
+	free_linkages(sent);
+	post_process_free(sent->postprocessor);
+	post_process_free(sent->constituent_pp);
 
 	global_rand_state = sent->rand_state;
 	xfree((char *) sent, sizeof(struct Sentence_s));
@@ -1143,16 +800,6 @@ int sentence_null_count(Sentence sent)
 	return sent->null_count;
 }
 
-int sentence_num_thin_linkages(Sentence sent)
-{
-	if (!sent) return 0;
-#ifdef USE_FAT_LINKAGES
-	return sent->num_thin_linkages;
-#else
-	return sent->num_valid_linkages;
-#endif /* USE_FAT_LINKAGES */
-}
-
 int sentence_num_linkages_found(Sentence sent)
 {
 	if (!sent) return 0;
@@ -1176,22 +823,9 @@ int sentence_num_violations(Sentence sent, LinkageIdx i)
 	if (!sent) return 0;
 
 	/* The sat solver (currently) fails to fill in link_info */
-	if (!sent->link_info) return 0;
+	if (!sent->lnkages) return 0;
 	if (sent->num_linkages_alloced <= i) return 0; /* bounds check */
-	return sent->link_info[i].N_violations;
-}
-
-int sentence_and_cost(Sentence sent, LinkageIdx i) {
-#ifdef USE_FAT_LINKAGES
-	if (!sent) return 0;
-
-	/* The sat solver (currently) fails to fill in link_info */
-	if (!sent->link_info) return 0;
-	if (sent->num_linkages_alloced <= i) return 0; /* bounds check */
-	return sent->link_info[i].and_cost;
-#else
-	return 0;
-#endif /* USE_FAT_LINKAGES */
+	return sent->lnkages[i].lifo.N_violations;
 }
 
 double sentence_disjunct_cost(Sentence sent, LinkageIdx i)
@@ -1199,9 +833,9 @@ double sentence_disjunct_cost(Sentence sent, LinkageIdx i)
 	if (!sent) return 0.0;
 
 	/* The sat solver (currently) fails to fill in link_info */
-	if (!sent->link_info) return 0.0;
+	if (!sent->lnkages) return 0.0;
 	if (sent->num_linkages_alloced <= i) return 0.0; /* bounds check */
-	return sent->link_info[i].disjunct_cost;
+	return sent->lnkages[i].lifo.disjunct_cost;
 }
 
 int sentence_link_cost(Sentence sent, LinkageIdx i)
@@ -1209,23 +843,17 @@ int sentence_link_cost(Sentence sent, LinkageIdx i)
 	if (!sent) return 0;
 
 	/* The sat solver (currently) fails to fill in link_info */
-	if (!sent->link_info) return 0;
+	if (!sent->lnkages) return 0;
 	if (sent->num_linkages_alloced <= i) return 0; /* bounds check */
-	return sent->link_info[i].link_cost;
+	return sent->lnkages[i].lifo.link_cost;
 }
 
 /* ============================================================== */
 /* A kind of morphism post-processing */
 
-static inline bool
-	is_AFFIXTYPE_STEM(char infix_mark, const char *a, size_t len)
-	{ return (SUBSCRIPT_MARK == a[len]) && (infix_mark == a[len+1]); }
 static inline bool
 	is_AFFIXTYPE_PREFIX(char infix_mark, const char *a, size_t len)
 	{ return infix_mark == a[len-1]; }
-static inline bool
-	is_AFFIXTYPE_SUFFIX(char infix_mark, const char *a)
-	{ return (infix_mark == a[0] && '\0' != a[1]); }
 static inline bool
 	is_AFFIXTYPE_EMPTY(char infix_mark, const char *a, size_t len)
 	{ return (0 == strcmp(a, EMPTY_WORD_MARK)); }
@@ -1261,12 +889,11 @@ static inline bool
 #define AFFIXTYPE_END		'b'	/* end of input word */
 
 /** return true if its good, else return false */
-bool sane_linkage_morphism(Sentence sent, size_t lk, Parse_Options opts)
+bool sane_linkage_morphism(Sentence sent, Linkage lkg, Parse_Options opts)
 {
 	size_t i;
-	Parse_info pi = sent->parse_info;
-	Dictionary afdict = sent->dict->affix_table; /* for INFIX_MARK only */
-	const char infix_mark = INFIX_MARK;
+	Dictionary afdict = sent->dict->affix_table;
+	const char infix_mark = INFIX_MARK(afdict);
 	int * matched_alts = NULL;       /* number of morphemes that have matched
 												 * the chosen disjuncts for the
 												 * unsplit_word, so far (index: ai) */
@@ -1283,10 +910,7 @@ bool sane_linkage_morphism(Sentence sent, size_t lk, Parse_Options opts)
 	char * affix_types_p = affix_types;
 	/* If all the words are null - behave as if everything matched. */
 	bool match_found = true;        /* djw matched a morpheme */
- 
-	Linkage_info * const lifo = &sent->link_info[lk];
 
-	extract_links(lifo->index, pi);
 	*affix_types_p = '\0';
 	for (i=0; i<sent->length; i++)
 	{
@@ -1295,14 +919,14 @@ bool sane_linkage_morphism(Sentence sent, size_t lk, Parse_Options opts)
 		size_t len;                /* disjunct length w/o subscript */
 		const char * mark;         /* char position of SUBSCRIPT_MARK */
 		bool empty_word = false;   /* is this an empty word? */
-		Disjunct * cdj = pi->chosen_disjuncts[i];
+		Disjunct * cdj = lkg->chosen_disjuncts[i];
 
-		lgdebug(+4, "Linkage %zu, word %zu/%zu\n", lk+1, i, sent->length);
+		lgdebug(+4, "Linkage %p, word %zu/%zu\n", lkg, i, sent->length);
 
 		/* Ignore island words */
 		if (NULL == cdj)
 		{
-			lgdebug(4, "%zu ignored island word\n", lk+1);
+			lgdebug(4, "%p ignored island word\n", lkg);
 			unsplit = NULL; /* mark it as island */
 			continue;
 		}
@@ -1320,7 +944,7 @@ bool sane_linkage_morphism(Sentence sent, size_t lk, Parse_Options opts)
 					realloc(matched_alts, sizeof(*matched_alts)*matched_alts_num);
 			}
 
-			lgdebug(4, "%zu unsplit word %s, alts:", lk+1, unsplit);
+			lgdebug(4, "%p unsplit word %s, alts:", lkg, unsplit);
 			for (ai = 0; ai < numalt; ai++)
 			{
 				matched_alts[ai] = 0;
@@ -1330,7 +954,7 @@ bool sane_linkage_morphism(Sentence sent, size_t lk, Parse_Options opts)
 		}
 		if (NULL == unsplit)
 		{
-			lgdebug(4, "%zu ignore morphemes of an island word\n", lk+1);
+			lgdebug(4, "%p ignore morphemes of an island word\n", lkg);
 			continue;
 		}
 
@@ -1346,12 +970,12 @@ bool sane_linkage_morphism(Sentence sent, size_t lk, Parse_Options opts)
 			empty_word = true;
 		}
 		else
-		if (is_AFFIXTYPE_SUFFIX(infix_mark, djw))
+		if (is_suffix(infix_mark, djw))
 		{
 			*affix_types_p = AFFIXTYPE_SUFFIX;
 		}
 		else
-		if (is_AFFIXTYPE_STEM(infix_mark, djw, len))
+		if (is_stem(djw))
 		{
 			*affix_types_p = AFFIXTYPE_STEM;
 		}
@@ -1365,12 +989,12 @@ bool sane_linkage_morphism(Sentence sent, size_t lk, Parse_Options opts)
 			*affix_types_p = AFFIXTYPE_WORD;
 		}
 
-		lgdebug(4, "%zu djw=%s affixtype=%c\n",
-		        lk+1, djw, empty_word ? 'E' : *affix_types_p);
+		lgdebug(4, "%p djw=%s affixtype=%c\n",
+		        lkg, djw, empty_word ? 'E' : *affix_types_p);
 
 		if (! empty_word) affix_types_p++;
 
-		lgdebug(4, "%zu djw %s matched alt#:", lk+1, djw);
+		lgdebug(4, "%p djw %s matched alt#:", lkg, djw);
 		match_found = false;
 
 		/* Compare the chosen word djw to the alternatives */
@@ -1427,7 +1051,7 @@ try_again:
 				{
 					downcase_utf8_str(downcased, a, MAX_WORD);
 					lgdebug(4, "\n");
-					lgdebug(4, "%zu downcasing %s>%s\n", lk+1, a, downcased);
+					lgdebug(4, "%p downcasing %s>%s\n", lkg, a, downcased);
 					s = downcased;
 					goto try_again;
 				}
@@ -1448,8 +1072,8 @@ try_again:
 		{
 			int num_morphemes = i - unsplit_i + 1;
 
-			lgdebug(4, "%zu end of input word, num_morphemes %d\n",
-					lk+1, num_morphemes);
+			lgdebug(4, "%p end of input word, num_morphemes %d\n",
+					lkg, num_morphemes);
 			*affix_types_p++ = AFFIXTYPE_END;
 
 			/* Make sure that there exists an alternative in
@@ -1468,10 +1092,10 @@ try_again:
 
 			if (! match_found)
 			{
-				lgdebug(4, "%zu morphemes are missing in this linkage\n", lk+1);
+				lgdebug(4, "%p morphemes are missing in this linkage\n", lkg);
 				break;
 			}
-			lgdebug(4, "%zu Perfect match\n", lk+1);
+			lgdebug(4, "%p Perfect match\n", lkg);
 		}
 	}
 	*affix_types_p = '\0';
@@ -1491,23 +1115,24 @@ try_again:
 			       "run with !bad and !verbosity=4 to debug\n", affix_types);
 	}
 	else
-		lgdebug(4, "%zu morpheme type combination '%s'\n", lk+1, affix_types);
+		lgdebug(4, "%p morpheme type combination '%s'\n", lkg, affix_types);
 
 	if (matched_alts) free(matched_alts);
 
 	if (match_found)
 	{
-		lgdebug(4, "%zu SUCCEEDED\n", lk+1);
+		lgdebug(4, "%p SUCCEEDED\n", lkg);
 		return true;
 	}
 	else
 	{
+		Linkage_info * const lifo = &lkg->lifo;
 		/* Oh no ... invalid morpheme combination! */
 		sent->num_valid_linkages --;
 		lifo->N_violations++;
 		lifo->pp_violation_msg = "Invalid morphism construction.";
 		if (!test_enabled("display-invalid-morphism")) lifo->discarded = true;
-		lgdebug(4, "%zu FAILED\n", lk+1);
+		lgdebug(4, "%p FAILED\n", lkg);
 		return false;
 	}
 }
@@ -1517,13 +1142,13 @@ static void sane_morphism(Sentence sent, Parse_Options opts)
 	size_t N_invalid_morphism = 0;
 	size_t lk;
 
-	for (lk = 0; lk < sent->num_linkages_post_processed; lk++)
+	for (lk = 0; lk < sent->num_linkages_alloced; lk++)
 	{
+		Linkage lkg = &sent->lnkages[lk];
 		/* Don't bother with linkages that already failed post-processing... */
-		Linkage_info * const lifo = &sent->link_info[lk];
-		if (0 != lifo->N_violations) continue;
+		if (0 != lkg->lifo.N_violations) continue;
 
-      if (!sane_linkage_morphism(sent, lk, opts))
+      if (!sane_linkage_morphism(sent, lkg, opts))
           N_invalid_morphism ++;
 	}
 
@@ -1531,10 +1156,11 @@ static void sane_morphism(Sentence sent, Parse_Options opts)
 	{
 		prt_error("Info: sane_morphism(): %zu of %zu linkages had "
 					 "invalid morphology construction\n",
-		          N_invalid_morphism, sent->num_valid_linkages);
+		          N_invalid_morphism, sent->num_linkages_alloced);
 	}
 }
 
+/** Misnamed, this has nothing to do with chart parsing */
 static void chart_parse(Sentence sent, Parse_Options opts)
 {
 	int nl;
@@ -1552,8 +1178,8 @@ static void chart_parse(Sentence sent, Parse_Options opts)
 
 	/* A parse set may have been already been built for this sentence,
 	 * if it was previously parsed.  If so we free it up before
-	 * building another.  */
-	if (sent->parse_info) free_parse_info(sent->parse_info);
+	 * building another.  Huh ?? How could that happen? */
+	free_parse_info(sent->parse_info);
 	sent->parse_info = parse_info_new(sent->length);
 
 	for (nl = opts->min_null_count; nl <= opts->max_null_count ; ++nl)
@@ -1576,18 +1202,11 @@ static void chart_parse(Sentence sent, Parse_Options opts)
 		sent->num_linkages_found = (int) total;
 		print_time(opts, "Counted parses");
 
-		if (test_enabled("old-pp-order"))
-		{
-			old_post_process_linkages(sent, mchxt, ctxt, opts);
-			sane_morphism(sent, opts);
-		}
-		else
-		{
-			select_linkages(sent, mchxt, ctxt, opts);
-			sane_morphism(sent, opts);
-			post_process_linkages(sent, opts);
-			sort_linkages(sent, opts);
-		}
+		select_linkages(sent, mchxt, ctxt, opts);
+		compute_chosen_disjuncts(sent);
+		sane_morphism(sent, opts);
+		post_process_linkages(sent, opts);
+		sort_linkages(sent, opts);
 		if (sent->num_valid_linkages > 0) break;
 
 		/* If we are here, then no valid linakges were found.
@@ -1640,18 +1259,11 @@ int sentence_parse(Sentence sent, Parse_Options opts)
 
 	/* Initialize/free any leftover garbage */
 	free_sentence_disjuncts(sent);  /* Is this really needed ??? */
-#ifdef USE_FAT_LINKAGES
-	if (sentence_contains_conjunction(sent)) free_AND_tables(sent);
-#endif /* USE_FAT_LINKAGES */
 	resources_reset_space(opts->resources);
 
 	if (resources_exhausted(opts->resources))
 		return 0;
 
-#ifdef USE_FAT_LINKAGES
-	init_analyze(sent);
-#endif /* USE_FAT_LINKAGES */
-
 	/* Expressions were previously set up during the tokenize stage. */
 	expression_prune(sent);
 	print_time(opts, "Finished expression pruning");
