@@ -1,6 +1,7 @@
 /*************************************************************************/
 /* Copyright (c) 2004                                                    */
 /* Daniel Sleator, David Temperley, and John Lafferty                    */
+/* Copyright (c) 2014 Linas Vepstas                                      */
 /* All rights reserved                                                   */
 /*                                                                       */
 /* Use of the link grammar parsing system is subject to the terms of the */
@@ -40,7 +41,7 @@
 
 bool post_process_match(const char *s, const char *t)
 {
-	char c;
+	if (NULL == t) return false;
 	if (islower((int)*t)) t++; /* Skip head-dependent indicator */
 	while (isupper((int)*s) || isupper((int)*t))
 	{
@@ -52,6 +53,7 @@ bool post_process_match(const char *s, const char *t)
 	{
 		if (*s != '#')
 		{
+			char c;
 			if (*t == '\0') c = '*'; else c = *t;
 			if (*s != c) return false;
 		}
@@ -91,9 +93,9 @@ static size_t find_domain_name(Postprocessor *pp, const char *link)
 
 /** Returns true if domain d1 is contained in domain d2 */
 static int contained_in(const Domain * d1, const Domain * d2,
-                        const Sublinkage *sublinkage)
+                        const Linkage sublinkage)
 {
-	bool mark[sublinkage->num_links];
+	bool *mark = alloca(sublinkage->num_links*sizeof(bool));
 	List_o_links * lol;
 	memset(mark, 0, sublinkage->num_links*(sizeof(bool)));
 	for (lol=d2->lol; lol != NULL; lol = lol->next)
@@ -129,14 +131,14 @@ static bool check_domain_nesting(Postprocessor *pp, int num_links)
 		for (d2=d1+1; d2 < pp->pp_data.domain_array + pp->pp_data.N_domains; d2++) {
 			memset(mark, 0, num_links*(sizeof mark[0]));
 			for (lol=d2->lol; lol != NULL; lol = lol->next) {
-		mark[lol->link] = 1;
+				mark[lol->link] = 1;
 			}
 			for (lol=d1->lol; lol != NULL; lol = lol->next) {
-		mark[lol->link] += 2;
+				mark[lol->link] += 2;
 			}
 			counts[0] = counts[1] = counts[2] = counts[3] = 0;
 			for (i=0; i<num_links; i++)
-		counts[(int)mark[i]]++;/* (int) cast avoids compiler warning DS 7/97 */
+				counts[(int)mark[i]]++;/* (int) cast avoids compiler warning DS 7/97 */
 			if ((counts[1] > 0) && (counts[2] > 0) && (counts[3] > 0))
 		return false;
 		}
@@ -155,7 +157,7 @@ static void free_List_o_links(List_o_links *lol)
 	while (lol != NULL)
 	{
 		xlol = lol->next;
-		xfree(lol, sizeof(List_o_links));
+		free(lol);
 		lol = xlol;
 	}
 }
@@ -166,7 +168,7 @@ static void free_D_tree_leaves(DTreeLeaf *dtl)
 	while (dtl != NULL)
 	{
 		xdtl = dtl->next;
-		xfree(dtl, sizeof(DTreeLeaf));
+		free(dtl);
 		dtl = xdtl;
 	}
 }
@@ -177,12 +179,13 @@ static void free_D_tree_leaves(DTreeLeaf *dtl)
 void post_process_free_data(PP_data * ppd)
 {
 	size_t w, d;
-	for (w = 0; w < ppd->length; w++)
+	for (w = 0; w < ppd->wowlen; w++)
 	{
 		free_List_o_links(ppd->word_links[w]);
 		ppd->word_links[w] = NULL;
 	}
-	for (d = 0; d < ppd->N_domains; d++)
+
+	for (d = 0; d < ppd->domlen; d++)
 	{
 		free_List_o_links(ppd->domain_array[d].lol);
 		ppd->domain_array[d].lol = NULL;
@@ -191,13 +194,16 @@ void post_process_free_data(PP_data * ppd)
 	}
 	free_List_o_links(ppd->links_to_ignore);
 	ppd->links_to_ignore = NULL;
+	ppd->length = 0;
+	ppd->N_domains = 0;
 }
 
 #ifdef THIS_FUNCTION_IS_NOT_CURRENTLY_USED
-static void connectivity_dfs(Postprocessor *pp, Sublinkage *sublinkage,
+static void connectivity_dfs(Postprocessor *pp, Linkage sublinkage,
                              int w, pp_linkset *ls)
 {
 	List_o_links *lol;
+	assert(w < pp->pp_data.length, "Bad word index");
 	pp->visited[w] = true;
 	for (lol = pp->pp_data.word_links[w]; lol != NULL; lol = lol->next)
 	{
@@ -212,10 +218,9 @@ static void chk_d_type(PP_node* ppn, size_t idx)
 {
 	if (ppn->dtsz <= idx)
 	{
-		size_t oldsz = ppn->dtsz;
-		ppn->dtsz = 2 * ppn->dtsz + 10;
-		ppn->d_type_array = xrealloc(ppn->d_type_array,
-			oldsz * sizeof(D_type_list*), ppn->dtsz * sizeof(D_type_list*));
+		ppn->dtsz += idx + 5;
+		ppn->d_type_array = realloc(ppn->d_type_array,
+			ppn->dtsz * sizeof(D_type_list*));
 	}
 }
 
@@ -229,7 +234,7 @@ static void build_type_array(Postprocessor *pp)
 		for (lol = pp->pp_data.domain_array[d].lol; lol != NULL; lol = lol->next)
 		{
 			chk_d_type(pp->pp_node, lol->link);
-			dtl = (D_type_list *) xalloc(sizeof(D_type_list));
+			dtl = (D_type_list *) malloc(sizeof(D_type_list));
 			dtl->next = pp->pp_node->d_type_array[lol->link];
 			dtl->type = pp->pp_data.domain_array[d].type;
 			pp->pp_node->d_type_array[lol->link] = dtl;
@@ -237,42 +242,16 @@ static void build_type_array(Postprocessor *pp)
 	}
 }
 
-#ifndef USE_FAT_LINKAGES
-static
-#endif /* USE_FAT_LINKAGES */
-void free_d_type(D_type_list * dtl)
+static void free_d_type(D_type_list * dtl)
 {
 	D_type_list * dtlx;
 	for (; dtl != NULL; dtl = dtlx)
 	{
 		dtlx = dtl->next;
-		xfree((void*) dtl, sizeof(D_type_list));
+		free((void*) dtl);
 	}
 }
 
-#ifdef USE_FAT_LINKAGES
-D_type_list * copy_d_type(const D_type_list * dtl)
-{
-	D_type_list *dtlx, *dtlcurr=NULL, *dtlhead=NULL;
-	for (; dtl!=NULL; dtl=dtl->next)
-	{
-		dtlx = (D_type_list *) xalloc(sizeof(D_type_list));
-		*dtlx = *dtl;
-		if (dtlhead == NULL)
-		{
-			dtlhead = dtlx;
-			dtlcurr = dtlx;
-		}
-		else
-		{
-			dtlcurr->next = dtlx;
-			dtlcurr = dtlx;
-		}
-	}
-	return dtlhead;
-}
-#endif /* USE_FAT_LINKAGES */
-
 /** free the pp node from last time */
 static void free_pp_node(Postprocessor *pp)
 {
@@ -285,18 +264,22 @@ static void free_pp_node(Postprocessor *pp)
 	{
 		free_d_type(ppn->d_type_array[i]);
 	}
-	xfree(ppn->d_type_array, ppn->dtsz * sizeof(D_type_list *));
-	xfree((void*) ppn, sizeof(PP_node));
+	free(ppn->d_type_array);
+	free((void*) ppn);
 }
 
 /** set up a fresh pp_node for later use */
 static void alloc_pp_node(Postprocessor *pp)
 {
-	PP_node *ppn = (PP_node *) xalloc(sizeof(PP_node));
+	size_t dz;
+	PP_node *ppn = (PP_node *) malloc(sizeof(PP_node));
 
 	/* highly unlikely that the number of links will ever exceed this */
 	ppn->dtsz = 2 * pp->pp_data.length;
-	ppn->d_type_array = (D_type_list **) xalloc (ppn->dtsz * sizeof(D_type_list*));
+	dz = ppn->dtsz * sizeof(D_type_list*);
+	ppn->d_type_array = (D_type_list **) malloc (dz);
+	memset(ppn->d_type_array, 0, dz);
+
 	pp->pp_node = ppn;
 }
 
@@ -308,15 +291,18 @@ static void clear_pp_node(Postprocessor *pp)
 
 	ppn->violation = NULL;
 	for (i=0; i<ppn->dtsz; i++)
+	{
+		free_d_type(ppn->d_type_array[i]);
 		ppn->d_type_array[i] = NULL;
+	}
 }
 
 
 /************************ rule application *******************************/
 
 static bool apply_rules(Postprocessor *pp,
-                        bool (applyfn) (Postprocessor *, Sublinkage *, pp_rule *),
-                        Sublinkage *sublinkage,
+                        bool (applyfn) (Postprocessor *, Linkage, pp_rule *),
+                        Linkage sublinkage,
                         pp_rule *rule_array,
                         const char **msg)
 {
@@ -334,8 +320,8 @@ static bool apply_rules(Postprocessor *pp,
 
 static bool
 apply_relevant_rules(Postprocessor *pp,
-                     bool (applyfn)(Postprocessor *, Sublinkage*, pp_rule *),
-                     Sublinkage *sublinkage,
+                     bool (applyfn)(Postprocessor *, Linkage, pp_rule *),
+                     Linkage sublinkage,
                      pp_rule *rule_array,
                      int *relevant_rules,
                      const char **msg)
@@ -363,7 +349,7 @@ apply_relevant_rules(Postprocessor *pp,
  * string matching)
  */
 static bool
-apply_contains_one(Postprocessor *pp, Sublinkage *sublinkage, pp_rule *rule)
+apply_contains_one(Postprocessor *pp, Linkage sublinkage, pp_rule *rule)
 {
 	DTreeLeaf * dtl;
 	size_t d, count;
@@ -372,7 +358,7 @@ apply_contains_one(Postprocessor *pp, Sublinkage *sublinkage, pp_rule *rule)
 		for (dtl = pp->pp_data.domain_array[d].child;
 		     dtl != NULL &&
 		        !post_process_match(rule->selector,
-		           sublinkage->link[dtl->link]->link_name);
+		           sublinkage->link_array[dtl->link].link_name);
 		     dtl = dtl->next) {}
 		if (dtl != NULL)
 		{
@@ -380,7 +366,7 @@ apply_contains_one(Postprocessor *pp, Sublinkage *sublinkage, pp_rule *rule)
 			count=0;
 			for (dtl = pp->pp_data.domain_array[d].child; dtl != NULL; dtl = dtl->next)
 			{
-				if (string_in_list(sublinkage->link[dtl->link]->link_name,
+				if (string_in_list(sublinkage->link_array[dtl->link].link_name,
 									 rule->link_array))
 				{
 					count=1;
@@ -400,7 +386,7 @@ apply_contains_one(Postprocessor *pp, Sublinkage *sublinkage, pp_rule *rule)
  * from the link_array contained in the rule. Uses exact string matching.
  */
 static bool
-apply_contains_none(Postprocessor *pp, Sublinkage *sublinkage, pp_rule *rule)
+apply_contains_none(Postprocessor *pp, Linkage sublinkage, pp_rule *rule)
 {
 	size_t d;
 	for (d=0; d<pp->pp_data.N_domains; d++)
@@ -409,14 +395,14 @@ apply_contains_none(Postprocessor *pp, Sublinkage *sublinkage, pp_rule *rule)
 		for (dtl = pp->pp_data.domain_array[d].child;
 		     dtl != NULL &&
 		         !post_process_match(rule->selector,
-		                  sublinkage->link[dtl->link]->link_name);
+		                  sublinkage->link_array[dtl->link].link_name);
 		     dtl = dtl->next) {}
 		if (dtl != NULL)
 		{
 			/* selector link of rule appears in this domain */
 			for (dtl = pp->pp_data.domain_array[d].child; dtl != NULL; dtl = dtl->next)
 			{
-				if (string_in_list(sublinkage->link[dtl->link]->link_name,
+				if (string_in_list(sublinkage->link_array[dtl->link].link_name,
 				                   rule->link_array))
 					return false;
 			}
@@ -431,17 +417,13 @@ apply_contains_none(Postprocessor *pp, Sublinkage *sublinkage, pp_rule *rule)
  * (2) it does, and it also contains one or more from the rule's link set
  */
 static bool
-apply_contains_one_globally(Postprocessor *pp, Sublinkage *sublinkage, pp_rule *rule)
+apply_contains_one_globally(Postprocessor *pp, Linkage sublinkage, pp_rule *rule)
 {
 	size_t i, j, count;
 	for (i = 0; i < sublinkage->num_links; i++)
 	{
-#ifdef USE_FAT_LINKAGES
-		if (sublinkage->link[i]->lw == SIZE_MAX) continue;
-#else
-		assert(sublinkage->link[i]->lw != SIZE_MAX);
-#endif
-		if (post_process_match(rule->selector,sublinkage->link[i]->link_name)) break;
+		assert(sublinkage->link_array[i].lw != SIZE_MAX);
+		if (post_process_match(rule->selector, sublinkage->link_array[i].link_name)) break;
 	}
 	if (i == sublinkage->num_links) return true;
 
@@ -449,12 +431,8 @@ apply_contains_one_globally(Postprocessor *pp, Sublinkage *sublinkage, pp_rule *
 	count = 0;
 	for (j = 0; j < sublinkage->num_links && count == 0; j++)
 	{
-#ifdef USE_FAT_LINKAGES
-		if (sublinkage->link[j]->lw == SIZE_MAX) continue;
-#else
-		assert(sublinkage->link[j]->lw != SIZE_MAX);
-#endif
-		if (string_in_list(sublinkage->link[j]->link_name, rule->link_array))
+		assert(sublinkage->link_array[j].lw != SIZE_MAX);
+		if (string_in_list(sublinkage->link_array[j].link_name, rule->link_array))
 		{
 			count = 1;
 			break;
@@ -470,14 +448,16 @@ apply_contains_one_globally(Postprocessor *pp, Sublinkage *sublinkage, pp_rule *
  * link are still connected.
  */
 static void reachable_without_dfs(Postprocessor *pp,
-                    Sublinkage *sublinkage, size_t a, size_t b, size_t w)
+                    Linkage sublinkage, size_t a, size_t b, size_t w)
 {
 	/* This is a depth first search of words reachable from w, excluding
 	 * any direct edge between word a and word b. */
 	List_o_links *lol;
+	assert(w < pp->pp_data.length, "Bad word index");
 	pp->visited[w] = true;
 	for (lol = pp->pp_data.word_links[w]; lol != NULL; lol = lol->next)
 	{
+		assert(lol->word < pp->pp_data.length, "Bad word index");
 		if (!pp->visited[lol->word] &&
 		    !(w == a && lol->word == b) &&
 		    !(w == b && lol->word == a))
@@ -495,7 +475,7 @@ static void reachable_without_dfs(Postprocessor *pp,
  * these links.
  */
 static bool
-apply_must_form_a_cycle(Postprocessor *pp, Sublinkage *sublinkage, pp_rule *rule)
+apply_must_form_a_cycle(Postprocessor *pp, Linkage sublinkage, pp_rule *rule)
 {
 	List_o_links *lol;
 	size_t w;
@@ -504,7 +484,8 @@ apply_must_form_a_cycle(Postprocessor *pp, Sublinkage *sublinkage, pp_rule *rule
 		for (lol = pp->pp_data.word_links[w]; lol != NULL; lol = lol->next)
 		{
 			if (w > lol->word) continue;	/* only consider each edge once */
-			if (!pp_linkset_match(rule->link_set, sublinkage->link[lol->link]->link_name)) continue;
+			if (!pp_linkset_match(rule->link_set, sublinkage->link_array[lol->link].link_name)) continue;
+
 			memset(pp->visited, 0, pp->pp_data.length*(sizeof pp->visited[0]));
 			reachable_without_dfs(pp, sublinkage, w, lol->word, w);
 			if (!pp->visited[lol->word]) return false;
@@ -513,11 +494,14 @@ apply_must_form_a_cycle(Postprocessor *pp, Sublinkage *sublinkage, pp_rule *rule
 
 	for (lol = pp->pp_data.links_to_ignore; lol != NULL; lol = lol->next)
 	{
-		w = sublinkage->link[lol->link]->lw;
+		w = sublinkage->link_array[lol->link].lw;
 		/* (w, lol->word) are the left and right ends of the edge we're considering */
-		if (!pp_linkset_match(rule->link_set, sublinkage->link[lol->link]->link_name)) continue;
+		if (!pp_linkset_match(rule->link_set, sublinkage->link_array[lol->link].link_name)) continue;
+
 		memset(pp->visited, 0, pp->pp_data.length*(sizeof pp->visited[0]));
 		reachable_without_dfs(pp, sublinkage, w, lol->word, w);
+
+		assert(lol->word < pp->pp_data.length, "Bad word index");
 		if (!pp->visited[lol->word]) return false;
 	}
 
@@ -530,7 +514,7 @@ apply_must_form_a_cycle(Postprocessor *pp, Sublinkage *sublinkage, pp_rule *rule
  * of the root word of the domain.
  */
 static bool
-apply_bounded(Postprocessor *pp, Sublinkage *sublinkage, pp_rule *rule)
+apply_bounded(Postprocessor *pp, Linkage sublinkage, pp_rule *rule)
 {
 	size_t d, lw;
 	char d_type;
@@ -539,10 +523,10 @@ apply_bounded(Postprocessor *pp, Sublinkage *sublinkage, pp_rule *rule)
 	for (d = 0; d < pp->pp_data.N_domains; d++)
 	{
 		if (pp->pp_data.domain_array[d].type != d_type) continue;
-		lw = sublinkage->link[pp->pp_data.domain_array[d].start_link]->lw;
+		lw = sublinkage->link_array[pp->pp_data.domain_array[d].start_link].lw;
 		for (lol = pp->pp_data.domain_array[d].lol; lol != NULL; lol = lol->next)
 		{
-			if (sublinkage->link[lol->link]->lw < lw) return false;
+			if (sublinkage->link_array[lol->link].lw < lw) return false;
 		}
 	}
 	return true;
@@ -551,53 +535,80 @@ apply_bounded(Postprocessor *pp, Sublinkage *sublinkage, pp_rule *rule)
 /**
  * fill in the pp->pp_data.word_links array with a list of words
  * neighboring each word (actually a list of links).	The dir fields
- * are not set, since this (after fat-link-extraction) is an
- * undirected graph.
+ * are not set, since this is an undirected graph.
  */
-static void build_graph(Postprocessor *pp, Sublinkage *sublinkage)
+static void build_graph(Postprocessor *pp, Linkage sublinkage)
 {
-	size_t i, link;
+	size_t link;
 	List_o_links * lol;
 
-	for (i = 0; i < pp->pp_data.length; i++)
-		pp->pp_data.word_links[i] = NULL;
+	/* Get more size, if needed */
+	if (pp->pp_data.wowlen <= pp->pp_data.length)
+	{
+		size_t newsz;
+		pp->pp_data.wowlen += pp->pp_data.length;
+		newsz = pp->pp_data.wowlen * sizeof(List_o_links *);
+		pp->pp_data.word_links = (List_o_links **) realloc(
+			pp->pp_data.word_links, newsz);
+	}
+	memset(pp->pp_data.word_links, 0, pp->pp_data.wowlen * sizeof(List_o_links *));
 
 	for (link = 0; link < sublinkage->num_links; link++)
 	{
-#ifdef USE_FAT_LINKAGES
-		if (sublinkage->link[link]->lw == SIZE_MAX) continue;
-#else
-		assert (sublinkage->link[link]->lw != SIZE_MAX);
-#endif
-		if (pp_linkset_match(pp->knowledge->ignore_these_links, sublinkage->link[link]->link_name))
+		assert (sublinkage->link_array[link].lw != SIZE_MAX);
+		if (NULL == sublinkage->link_array[link].link_name) continue;
+		if (pp_linkset_match(pp->knowledge->ignore_these_links,
+		                     sublinkage->link_array[link].link_name))
 		{
-			lol = (List_o_links *) xalloc(sizeof(List_o_links));
+			lol = (List_o_links *) malloc(sizeof(List_o_links));
 			lol->next = pp->pp_data.links_to_ignore;
 			pp->pp_data.links_to_ignore = lol;
 			lol->link = link;
-			lol->word = sublinkage->link[link]->rw;
+			lol->word = sublinkage->link_array[link].rw;
 			continue;
 		}
 
-		lol = (List_o_links *) xalloc(sizeof(List_o_links));
-		lol->next = pp->pp_data.word_links[sublinkage->link[link]->lw];
-		pp->pp_data.word_links[sublinkage->link[link]->lw] = lol;
+		lol = (List_o_links *) malloc(sizeof(List_o_links));
+		lol->next = pp->pp_data.word_links[sublinkage->link_array[link].lw];
+		pp->pp_data.word_links[sublinkage->link_array[link].lw] = lol;
 		lol->link = link;
-		lol->word = sublinkage->link[link]->rw;
+		lol->word = sublinkage->link_array[link].rw;
 
-		lol = (List_o_links *) xalloc(sizeof(List_o_links));
-		lol->next = pp->pp_data.word_links[sublinkage->link[link]->rw];
-		pp->pp_data.word_links[sublinkage->link[link]->rw] = lol;
+		lol = (List_o_links *) malloc(sizeof(List_o_links));
+		lol->next = pp->pp_data.word_links[sublinkage->link_array[link].rw];
+		pp->pp_data.word_links[sublinkage->link_array[link].rw] = lol;
 		lol->link = link;
-		lol->word = sublinkage->link[link]->lw;
+		lol->word = sublinkage->link_array[link].lw;
 	}
 }
 
-static void setup_domain_array(Postprocessor *pp,
-                               int n, const char *string, int start_link)
+static void setup_domain_array(Postprocessor *pp, size_t n,
+                               const char *string, int start_link)
 {
 	/* set pp->visited[i] to false */
+	/* Grab more memory if needed */
+	if (pp->vlength <= pp->pp_data.length)
+	{
+		size_t newsz;
+		pp->vlength += pp->pp_data.length;
+		newsz = pp->vlength * sizeof(bool);
+		pp->visited = (bool *) realloc(pp->visited, newsz);
+	}
 	memset(pp->visited, 0, pp->pp_data.length*(sizeof pp->visited[0]));
+
+	/* Grab more memory if needed */
+	if (pp->pp_data.domlen <= n)
+	{
+		size_t newsz, incsz;
+#define DOMINC 16
+		pp->pp_data.domlen += DOMINC;
+		newsz = pp->pp_data.domlen * sizeof(Domain);
+		incsz = DOMINC * sizeof(Domain);
+		pp->pp_data.domain_array = (Domain *) realloc(pp->pp_data.domain_array,
+			newsz);
+		memset(&pp->pp_data.domain_array[pp->pp_data.domlen], 0, incsz);
+	}
+
 	pp->pp_data.domain_array[n].string = string;
 	pp->pp_data.domain_array[n].lol    = NULL;
 	pp->pp_data.domain_array[n].size   = 0;
@@ -607,17 +618,18 @@ static void setup_domain_array(Postprocessor *pp,
 static void add_link_to_domain(Postprocessor *pp, int link)
 {
 	List_o_links *lol;
-	lol = (List_o_links *) xalloc(sizeof(List_o_links));
+	lol = (List_o_links *) malloc(sizeof(List_o_links));
 	lol->next = pp->pp_data.domain_array[pp->pp_data.N_domains].lol;
 	pp->pp_data.domain_array[pp->pp_data.N_domains].lol = lol;
 	pp->pp_data.domain_array[pp->pp_data.N_domains].size++;
 	lol->link = link;
 }
 
-static void depth_first_search(Postprocessor *pp, Sublinkage *sublinkage,
+static void depth_first_search(Postprocessor *pp, Linkage sublinkage,
                                size_t w, size_t root, size_t start_link)
 {
 	List_o_links *lol;
+	assert(w < pp->pp_data.length, "Bad word index");
 	pp->visited[w] = true;
 	for (lol = pp->pp_data.word_links[w]; lol != NULL; lol = lol->next)
 	{
@@ -631,17 +643,18 @@ static void depth_first_search(Postprocessor *pp, Sublinkage *sublinkage,
 		if (!pp->visited[lol->word] && (lol->word != root) &&
 		       !(lol->word < root && lol->word < w &&
 		       pp_linkset_match(pp->knowledge->restricted_links,
-		                sublinkage->link[lol->link]->link_name)))
+		                sublinkage->link_array[lol->link].link_name)))
 		{
 			depth_first_search(pp, sublinkage, lol->word, root, start_link);
 		}
 	}
 }
 
-static void bad_depth_first_search(Postprocessor *pp, Sublinkage *sublinkage,
+static void bad_depth_first_search(Postprocessor *pp, Linkage sublinkage,
                                    size_t w, size_t root, size_t start_link)
 {
 	List_o_links * lol;
+	assert(w < pp->pp_data.length, "Bad word index");
 	pp->visited[w] = true;
 	for (lol = pp->pp_data.word_links[w]; lol != NULL; lol = lol->next)
 	{
@@ -652,20 +665,22 @@ static void bad_depth_first_search(Postprocessor *pp, Sublinkage *sublinkage,
 	}
 	for (lol = pp->pp_data.word_links[w]; lol != NULL; lol = lol->next)
 	{
+		assert(lol->word < pp->pp_data.length, "Bad word index");
 		if ((!pp->visited[lol->word]) && !(w == root && lol->word < w) &&
 		     !(lol->word < root && lol->word < w &&
 		          pp_linkset_match(pp->knowledge->restricted_links,
-		                sublinkage->link[lol->link]->link_name)))
+		                sublinkage->link_array[lol->link].link_name)))
 		{
 			bad_depth_first_search(pp, sublinkage, lol->word, root, start_link);
 		}
 	}
 }
 
-static void d_depth_first_search(Postprocessor *pp, Sublinkage *sublinkage,
+static void d_depth_first_search(Postprocessor *pp, Linkage sublinkage,
                     size_t w, size_t root, size_t right, size_t start_link)
 {
 	List_o_links * lol;
+	assert(w < pp->pp_data.length, "Bad word index");
 	pp->visited[w] = true;
 	for (lol = pp->pp_data.word_links[w]; lol != NULL; lol = lol->next)
 	{
@@ -676,21 +691,23 @@ static void d_depth_first_search(Postprocessor *pp, Sublinkage *sublinkage,
 	}
 	for (lol = pp->pp_data.word_links[w]; lol != NULL; lol = lol->next)
 	{
+		assert(lol->word < pp->pp_data.length, "Bad word index");
 		if (!pp->visited[lol->word] && !(w == root && lol->word >= right) &&
 		    !(w == root && lol->word < root) &&
 		       !(lol->word < root && lol->word < w &&
 		          pp_linkset_match(pp->knowledge->restricted_links,
-		                   sublinkage->link[lol->link]->link_name)))
+		                   sublinkage->link_array[lol->link].link_name)))
 		{
 			d_depth_first_search(pp,sublinkage,lol->word,root,right,start_link);
 		}
 	}
 }
 
-static void left_depth_first_search(Postprocessor *pp, Sublinkage *sublinkage,
+static void left_depth_first_search(Postprocessor *pp, Linkage sublinkage,
                                     size_t w, size_t right, size_t start_link)
 {
 	List_o_links *lol;
+	assert(w < pp->pp_data.length, "Bad word index");
 	pp->visited[w] = true;
 	for (lol = pp->pp_data.word_links[w]; lol != NULL; lol = lol->next)
 	{
@@ -701,6 +718,7 @@ static void left_depth_first_search(Postprocessor *pp, Sublinkage *sublinkage,
 	}
 	for (lol = pp->pp_data.word_links[w]; lol != NULL; lol = lol->next)
 	{
+		assert(lol->word < pp->pp_data.length, "Bad word index");
 		if (!pp->visited[lol->word] && (lol->word != right))
 		{
 			depth_first_search(pp, sublinkage, lol->word, right, start_link);
@@ -713,7 +731,7 @@ static int domain_compare(const Domain * d1, const Domain * d2)
 	return (d1->size - d2->size); /* for sorting the domains by size */
 }
 
-static void build_domains(Postprocessor *pp, Sublinkage *sublinkage)
+static void build_domains(Postprocessor *pp, Linkage sublinkage)
 {
 	size_t link, i, d;
 	const char *s;
@@ -721,12 +739,9 @@ static void build_domains(Postprocessor *pp, Sublinkage *sublinkage)
 
 	for (link = 0; link<sublinkage->num_links; link++)
 	{
-#ifdef USE_FAT_LINKAGES
-		if (sublinkage->link[link]->lw == SIZE_MAX) continue;
-#else
-		assert (sublinkage->link[link]->lw != SIZE_MAX);
-#endif
-		s = sublinkage->link[link]->link_name;
+		assert (sublinkage->link_array[link].lw != SIZE_MAX);
+		if (NULL == sublinkage->link_array[link].link_name) continue;
+		s = sublinkage->link_array[link].link_name;
 
 		if (pp_linkset_match(pp->knowledge->ignore_these_links, s)) continue;
 		if (pp_linkset_match(pp->knowledge->domain_starter_links, s))
@@ -734,31 +749,31 @@ static void build_domains(Postprocessor *pp, Sublinkage *sublinkage)
 			setup_domain_array(pp, pp->pp_data.N_domains, s, link);
 			if (pp_linkset_match(pp->knowledge->domain_contains_links, s))
 				add_link_to_domain(pp, link);
-			depth_first_search(pp,sublinkage,sublinkage->link[link]->rw,
-							 sublinkage->link[link]->lw, link);
+			depth_first_search(pp, sublinkage, sublinkage->link_array[link].rw,
+							 sublinkage->link_array[link].lw, link);
 
 			pp->pp_data.N_domains++;
 			assert(pp->pp_data.N_domains<PP_MAX_DOMAINS, "raise value of PP_MAX_DOMAINS");
 		}
 		else
-		if (pp_linkset_match(pp->knowledge->urfl_domain_starter_links,s))
+		if (pp_linkset_match(pp->knowledge->urfl_domain_starter_links, s))
 		{
 			setup_domain_array(pp, pp->pp_data.N_domains, s, link);
 			/* always add the starter link to its urfl domain */
 			add_link_to_domain(pp, link);
-			bad_depth_first_search(pp,sublinkage,sublinkage->link[link]->rw,
-			                       sublinkage->link[link]->lw, link);
+			bad_depth_first_search(pp, sublinkage,sublinkage->link_array[link].rw,
+			                       sublinkage->link_array[link].lw, link);
 			pp->pp_data.N_domains++;
 			assert(pp->pp_data.N_domains<PP_MAX_DOMAINS,"raise PP_MAX_DOMAINS value");
 		}
 		else
-		if (pp_linkset_match(pp->knowledge->urfl_only_domain_starter_links,s))
+		if (pp_linkset_match(pp->knowledge->urfl_only_domain_starter_links, s))
 		{
 			setup_domain_array(pp, pp->pp_data.N_domains, s, link);
 			/* do not add the starter link to its urfl_only domain */
-			d_depth_first_search(pp,sublinkage, sublinkage->link[link]->lw,
-			                     sublinkage->link[link]->lw,
-			                     sublinkage->link[link]->rw, link);
+			d_depth_first_search(pp, sublinkage, sublinkage->link_array[link].lw,
+			                     sublinkage->link_array[link].lw,
+			                     sublinkage->link_array[link].rw, link);
 			pp->pp_data.N_domains++;
 			assert(pp->pp_data.N_domains<PP_MAX_DOMAINS,"raise PP_MAX_DOMAINS value");
 		}
@@ -767,8 +782,8 @@ static void build_domains(Postprocessor *pp, Sublinkage *sublinkage)
 		{
 			setup_domain_array(pp, pp->pp_data.N_domains, s, link);
 			/* do not add the starter link to a left domain */
-			left_depth_first_search(pp,sublinkage, sublinkage->link[link]->lw,
-			                        sublinkage->link[link]->rw, link);
+			left_depth_first_search(pp, sublinkage, sublinkage->link_array[link].lw,
+			                        sublinkage->link_array[link].rw, link);
 			pp->pp_data.N_domains++;
 			assert(pp->pp_data.N_domains<PP_MAX_DOMAINS,"raise PP_MAX_DOMAINS value");
 		}
@@ -791,7 +806,7 @@ static void build_domains(Postprocessor *pp, Sublinkage *sublinkage)
 	}
 }
 
-static void build_domain_forest(Postprocessor *pp, Sublinkage *sublinkage)
+static void build_domain_forest(Postprocessor *pp, Linkage sublinkage)
 {
 	size_t d, d1, link;
 	DTreeLeaf * dtl;
@@ -823,16 +838,12 @@ static void build_domain_forest(Postprocessor *pp, Sublinkage *sublinkage)
 	}
 	for (link=0; link < sublinkage->num_links; link++)
 	{
-#ifdef USE_FAT_LINKAGES
-		if (sublinkage->link[link]->lw == SIZE_MAX) continue;
-#else
-		assert (sublinkage->link[link]->lw != SIZE_MAX);
-#endif
+		assert (sublinkage->link_array[link].lw != SIZE_MAX);
 		for (d=0; d<pp->pp_data.N_domains; d++)
 		{
 			if (link_in_domain(link, &pp->pp_data.domain_array[d]))
 			{
-				dtl = (DTreeLeaf *) xalloc(sizeof(DTreeLeaf));
+				dtl = (DTreeLeaf *) malloc(sizeof(DTreeLeaf));
 				dtl->link = link;
 				dtl->parent = &pp->pp_data.domain_array[d];
 				dtl->next = pp->pp_data.domain_array[d].child;
@@ -844,16 +855,16 @@ static void build_domain_forest(Postprocessor *pp, Sublinkage *sublinkage)
 }
 
 static int
-internal_process(Postprocessor *pp, Sublinkage *sublinkage, const char **msg)
+internal_process(Postprocessor *pp, Linkage sublinkage, const char **msg)
 {
 	size_t i;
 	/* quick test: try applying just the relevant global rules */
-	if (!apply_relevant_rules(pp,apply_contains_one_globally,
+	if (!apply_relevant_rules(pp, apply_contains_one_globally,
 	                          sublinkage,
 	                          pp->knowledge->contains_one_rules,
 	                          pp->relevant_contains_one_rules, msg))
 	{
-		for (i = 0; i < pp->pp_data.length; i++)
+		for (i = 0; i < pp->pp_data.wowlen; i++)
 			pp->pp_data.word_links[i] = NULL;
 		pp->pp_data.N_domains = 0;
 		return -1;
@@ -938,65 +949,79 @@ static void prune_irrelevant_rules(Postprocessor *pp)
 
 /***************** definitions of exported functions ***********************/
 
+PostProcessor * post_process_open(const char *path)
+{
+	pp_knowledge *kno = pp_knowledge_open(path);
+	return post_process_new(kno);
+}
+
+void post_process_close(PostProcessor *pp)
+{
+	pp_knowledge_close(pp->knowledge);
+	post_process_free(pp);
+}
+
 /**
  * read rules from path and initialize the appropriate fields in
  * a postprocessor structure, a pointer to which is returned.
  */
-Postprocessor * post_process_open(const char *path)
+Postprocessor * post_process_new(pp_knowledge * kno)
 {
 	Postprocessor *pp;
-	if (path == NULL) return NULL;
 
-	pp = (Postprocessor *) xalloc (sizeof(Postprocessor));
-	pp->knowledge	= pp_knowledge_open(path);
-	pp->sentence_link_name_set = string_set_create();
+	pp = (Postprocessor *) malloc (sizeof(Postprocessor));
+	pp->knowledge = kno;
+	pp->string_set = string_set_create();
 	pp->set_of_links_of_sentence = pp_linkset_open(1024);
 	pp->set_of_links_in_an_active_rule = pp_linkset_open(1024);
 	pp->relevant_contains_one_rules =
-	      (int *) xalloc ((pp->knowledge->n_contains_one_rules + 1)
+	      (int *) malloc ((pp->knowledge->n_contains_one_rules + 1)
 	                      *(sizeof pp->relevant_contains_one_rules[0]));
 	pp->relevant_contains_none_rules =
-	      (int *) xalloc ((pp->knowledge->n_contains_none_rules + 1)
+	      (int *) malloc ((pp->knowledge->n_contains_none_rules + 1)
 	                      *(sizeof pp->relevant_contains_none_rules[0]));
 	pp->relevant_contains_one_rules[0]	= -1;
 	pp->relevant_contains_none_rules[0] = -1;
 	pp->pp_node = NULL;
-	pp->pp_data.links_to_ignore = NULL;
 	pp->n_local_rules_firing	= 0;
 	pp->n_global_rules_firing = 0;
+
+	pp->q_pruned_rules = false;
+
+	/* 60 is just starting size, these are expanded if needed */
+	pp->vlength = 60;
+	pp->visited = (bool*) malloc(pp->vlength * sizeof(bool));
+
+	pp->pp_data.links_to_ignore = NULL;
+	pp->pp_data.domlen = 60;
+	pp->pp_data.domain_array = (Domain*) malloc(pp->pp_data.domlen * sizeof(Domain));
+	memset(pp->pp_data.domain_array, 0, pp->pp_data.domlen * sizeof(Domain));
+
+	pp->pp_data.wowlen = 60;
+	pp->pp_data.word_links = (List_o_links **) malloc(pp->pp_data.wowlen * sizeof(List_o_links*));
+	memset(pp->pp_data.word_links, 0, pp->pp_data.wowlen * sizeof(List_o_links *));
+
 	return pp;
 }
 
-void post_process_close(Postprocessor *pp)
+void post_process_free(Postprocessor *pp)
 {
 	/* frees up memory associated with pp, previously allocated by open */
 	if (pp == NULL) return;
-	string_set_delete(pp->sentence_link_name_set);
+	string_set_delete(pp->string_set);
 	pp_linkset_close(pp->set_of_links_of_sentence);
 	pp_linkset_close(pp->set_of_links_in_an_active_rule);
-	xfree(pp->relevant_contains_one_rules,
-	      (1 + pp->knowledge->n_contains_one_rules)
-	      *(sizeof pp->relevant_contains_one_rules[0]));
-	xfree(pp->relevant_contains_none_rules,
-	      (1 + pp->knowledge->n_contains_none_rules)
-	      *(sizeof pp->relevant_contains_none_rules[0]));
-	pp_knowledge_close(pp->knowledge);
+	free(pp->relevant_contains_one_rules);
+	free(pp->relevant_contains_none_rules);
+	pp->knowledge = NULL;
 	free_pp_node(pp);
-	xfree(pp, sizeof(Postprocessor));
-}
 
-void post_process_close_sentence(Postprocessor *pp)
-{
-	if (pp == NULL) return;
-	pp_linkset_clear(pp->set_of_links_of_sentence);
-	pp_linkset_clear(pp->set_of_links_in_an_active_rule);
-	string_set_delete(pp->sentence_link_name_set);
-	pp->sentence_link_name_set = string_set_create();
-	pp->n_local_rules_firing	= 0;
-	pp->n_global_rules_firing = 0;
-	pp->relevant_contains_one_rules[0]	= -1;
-	pp->relevant_contains_none_rules[0] = -1;
-	free_pp_node(pp);
+	free(pp->visited);
+
+	post_process_free_data(&pp->pp_data);
+	free(pp->pp_data.domain_array);
+	free(pp->pp_data.word_links);
+	free(pp);
 }
 
 /**
@@ -1005,17 +1030,18 @@ void post_process_close_sentence(Postprocessor *pp)
  * simply maintain a set of "seen" link names for rule pruning later on
  */
 void post_process_scan_linkage(Postprocessor *pp, Parse_Options opts,
-									 Sentence sent, Sublinkage *sublinkage)
+                               Linkage sublinkage)
 {
-	const char *p;
 	size_t i;
 	if (pp == NULL) return;
-	if (sent->length < opts->twopass_length) return;
+	if (sublinkage->num_words < opts->twopass_length) return;
 	for (i = 0; i < sublinkage->num_links; i++)
 	{
-		/* if (sublinkage->link[i]->lw == SIZE_MAX) continue; */
-		assert(sublinkage->link[i]->lw != SIZE_MAX);
-		p = string_set_add(sublinkage->link[i]->link_name, pp->sentence_link_name_set);
+		const char *p;
+		assert(sublinkage->link_array[i].lw != SIZE_MAX);
+
+		p = string_set_add(sublinkage->link_array[i].link_name,
+		                   pp->string_set);
 		pp_linkset_add(pp->set_of_links_of_sentence, p);
 	}
 }
@@ -1079,21 +1105,22 @@ static void report_pp_stats(Postprocessor *pp)
 }
 
 /**
- * Takes a sublinkage and returns:
+ * Takes a linkage and returns:
  *  . for each link, the domain structure of that link
  *  . a list of the violation strings
- * NB: sublinkage->link[i]->l=-1 means that this connector
- * is to be ignored
+ * NB: linkage->link[i]->l=-1 means that this connector is to be ignored.
  */
 PP_node *do_post_process(Postprocessor *pp, Parse_Options opts,
-                         Sentence sent, Sublinkage *sublinkage, bool cleanup)
+                         Linkage sublinkage)
 {
 	const char *msg;
 
 	if (pp == NULL) return NULL;
 
+	// XXX wtf .. why is this not leaking memory ?
 	pp->pp_data.links_to_ignore = NULL;
-	pp->pp_data.length = sent->length;
+
+	pp->pp_data.length = sublinkage->num_words;
 
 	/* In the name of responsible memory management, we retain a copy of the
 	 * returned data structure pp_node as a field in pp, so that we can clear
@@ -1102,11 +1129,11 @@ PP_node *do_post_process(Postprocessor *pp, Parse_Options opts,
 
 	/* The first time we see a sentence, prune the rules which we won't be
 	 * needing during postprocessing the linkages of this sentence */
-	if (sent->q_pruned_rules == false && sent->length >= opts->twopass_length)
+	if (pp->q_pruned_rules == false && sublinkage->num_words >= opts->twopass_length)
 	{
 		prune_irrelevant_rules(pp);
 	}
-	sent->q_pruned_rules = true;
+	pp->q_pruned_rules = true;
 
 	switch (internal_process(pp, sublinkage, &msg))
 	{
@@ -1129,12 +1156,8 @@ PP_node *do_post_process(Postprocessor *pp, Parse_Options opts,
 	}
 
 	report_pp_stats(pp);
-
 	build_type_array(pp);
-	if (cleanup)
-	{
-		post_process_free_data(&pp->pp_data);
-	}
+
 	return pp->pp_node;
 }
 
