@@ -58,39 +58,36 @@ typedef struct
 {
 	int N_rows; /* N_rows -- the number of rows */
 	/* tells the height of the links above the sentence */
-	int link_heights[MAX_LINKS];
+	int * link_heights;
 	/* the word beginning each row of the display */
-	int row_starts[MAX_SENTENCE];
+	int * row_starts;
 } ps_ctxt_t;
 
 /**
- * prints s then prints the last |t|-|s| characters of t.
+ * Prints s then prints the last |t|-|s| characters of t.
  * if s is longer than t, it truncates s.
- * multi-byte enabled.
+ * Handles utf8 strings correctly.
  */
 static void left_append_string(String * string, const char * s, const char * t)
 {
 	size_t i;
-	size_t slen = strlen(s);
-	size_t tlen = strlen(t);
+	size_t slen = utf8_strlen(s);
+	size_t tlen = utf8_strlen(t);
 
-	for (i = 0; i < tlen; )
+	for (i = 0; i < tlen; i++)
 	{
 		if (i < slen)
-		{
-			i += append_utf8_char(string, &s[i]);
-		}
+			append_utf8_char(string, s);
 		else
-		{
-			i += append_utf8_char(string, &t[i]);
-		}
+			append_utf8_char(string, t);
+
+		s += utf8_next(s);
+		t += utf8_next(t);
 	}
 }
 
 static void print_a_link(String * s, const Linkage linkage, LinkIdx link)
 {
-	Sentence sent = linkage_get_sentence(linkage);
-	Dictionary dict = sent->dict;
 	WordIdx l, r;
 	const char *label, *llabel, *rlabel;
 
@@ -100,12 +97,11 @@ static void print_a_link(String * s, const Linkage linkage, LinkIdx link)
 	llabel = linkage_get_link_llabel(linkage, link);
 	rlabel = linkage_get_link_rlabel(linkage, link);
 
-	if ((l == 0) && dict->left_wall_defined)
+	if (l == 0)
 	{
 		left_append_string(s, LEFT_WALL_DISPLAY, "               ");
 	}
-	else if ((l == (linkage_get_num_words(linkage)-1)) &&
-	         dict->right_wall_defined)
+	else if (l == (linkage_get_num_words(linkage) - 1))
 	{
 		left_append_string(s, RIGHT_WALL_DISPLAY, "               ");
 	}
@@ -227,8 +223,7 @@ char * linkage_print_disjuncts(const Linkage linkage)
 	char * djs, *mark;
 	int w;
 	String * s = string_new();
-	Sentence sent = linkage->sent;
-	int nwords = sent->length;
+	int nwords = linkage->num_words;
 
 	/* Loop over each word in the sentence */
 	for (w = 0; w < nwords; w++)
@@ -236,7 +231,7 @@ char * linkage_print_disjuncts(const Linkage linkage)
 		int pad = 21;
 		double cost;
 		char infword[MAX_WORD];
-		Disjunct *disj = linkage->sent->parse_info->chosen_disjuncts[w];
+		Disjunct *disj = linkage->chosen_disjuncts[w];
 		if (NULL == disj) continue;
 
 		/* Cleanup the subscript mark before printing. */
@@ -273,58 +268,44 @@ build_linkage_postscript_string(const Linkage linkage,
 	int link, i,j;
 	int d;
 	int N_wall_connectors;
-	bool print_word_0 = false, print_word_N = false;
+	bool print_word_0, print_word_N;
 	bool suppressor_used;
-#ifdef USE_FAT_LINKAGES
-	Sublinkage *sublinkage = &(linkage->sublinkage[linkage->current]);
-#else
-	Sublinkage *sublinkage = &(linkage->sublinkage);
-#endif /* USE_FAT_LINKAGES */
-	int N_links = sublinkage->num_links;
-	Link **ppla = sublinkage->link;
+	int N_links = linkage->num_links;
+	Link *ppla = linkage->link_array;
 	String  * string;
 	char * ps_string;
-	Dictionary dict = linkage->sent->dict;
 	int N_words_to_print;
 
 	string = string_new();
 
 	N_wall_connectors = 0;
-	if (dict->left_wall_defined) {
-		suppressor_used = false;
-		if (!display_walls)
-			for (j=0; j<N_links; j++) {
-				if (ppla[j]->lw == 0) {
-					if (ppla[j]->rw == linkage->num_words-1) continue;
-					N_wall_connectors ++;
-					if (strcmp(ppla[j]->lc->string, LEFT_WALL_SUPPRESS)==0) {
-						suppressor_used = true;
-					}
-				}
-			}
-		print_word_0 = (((!suppressor_used) && (N_wall_connectors != 0))
-						|| (N_wall_connectors > 1) || display_walls);
-	} else {
-		print_word_0 = true;
-	}
-
-	N_wall_connectors = 0;
-	if (dict->right_wall_defined) {
-		suppressor_used = false;
+	suppressor_used = false;
+	if (!display_walls) {
 		for (j=0; j<N_links; j++) {
-			if (ppla[j]->rw == linkage->num_words-1) {
+			if (ppla[j].lw == 0) {
+				if (ppla[j].rw == linkage->num_words-1) continue;
 				N_wall_connectors ++;
-				if (strcmp(ppla[j]->lc->string, RIGHT_WALL_SUPPRESS)==0){
+				if (strcmp(ppla[j].lc->string, LEFT_WALL_SUPPRESS)==0) {
 					suppressor_used = true;
 				}
 			}
 		}
-		print_word_N = (((!suppressor_used) && (N_wall_connectors != 0))
-						|| (N_wall_connectors > 1) || display_walls);
 	}
-	else {
-		print_word_N = true;
+	print_word_0 = (((!suppressor_used) && (N_wall_connectors != 0))
+					|| (N_wall_connectors != 1) || display_walls);
+
+	N_wall_connectors = 0;
+	suppressor_used = false;
+	for (j=0; j<N_links; j++) {
+		if (ppla[j].rw == linkage->num_words-1) {
+			N_wall_connectors ++;
+			if (strcmp(ppla[j].lc->string, RIGHT_WALL_SUPPRESS)==0){
+				suppressor_used = true;
+			}
+		}
 	}
+	print_word_N = (((!suppressor_used) && (N_wall_connectors != 0))
+					|| (N_wall_connectors != 1) || display_walls);
 
 	if (print_word_0) d=0; else d=1;
 
@@ -344,25 +325,16 @@ build_linkage_postscript_string(const Linkage linkage,
 	append_string(string,"[");
 	j = 0;
 	for (link=0; link<N_links; link++) {
-		if (!print_word_0 && (ppla[link]->lw == 0)) continue;
-		if (!print_word_N && (ppla[link]->rw == linkage->num_words-1)) continue;
+		if (!print_word_0 && (ppla[link].lw == 0)) continue;
+		if (!print_word_N && (ppla[link].rw == linkage->num_words-1)) continue;
 		// if (ppla[link]->lw == SIZE_MAX) continue;
-		assert (ppla[link]->lw != SIZE_MAX);
+		assert (ppla[link].lw != SIZE_MAX);
 		if ((j%7 == 0) && (j>0)) append_string(string,"\n");
 		j++;
 		append_string(string,"[%zu %zu %d",
-				ppla[link]->lw - d, ppla[link]->rw - d,
+				ppla[link].lw - d, ppla[link].rw - d,
 				pctx->link_heights[link]);
-#ifdef USE_FAT_LINKAGES
-		if (ppla[link]->lc->label < 0) {
-			append_string(string," (%s)]", ppla[link]->link_name);
-		} else {
-			append_string(string," ()]");
-		}
-#else
-		append_string(string," (%s)]", ppla[link]->link_name);
-#endif /* USE_FAT_LINKAGES */
-
+		append_string(string," (%s)]", ppla[link].link_name);
 	}
 	append_string(string,"]");
 	append_string(string,"\n");
@@ -398,24 +370,18 @@ linkage_print_diagram_ctxt(const Linkage linkage,
                            ps_ctxt_t *pctx)
 {
 	bool display_short = true;
-	unsigned int i, j, k, cl, cr, row, top_row, top_row_p1;
+	unsigned int i, j, k, cl, cr, inc, row, top_row, top_row_p1;
 	const char *s;
 	char *t;
-	bool print_word_0 = false, print_word_N = false;
+	bool print_word_0 , print_word_N;
 	int N_wall_connectors;
 	bool suppressor_used;
-	int center[linkage->num_words+1];
+	int *center = alloca((linkage->num_words+1)*sizeof(int));
 	unsigned int line_len, link_length;
-#ifdef USE_FAT_LINKAGES
-	Sublinkage *sublinkage = &(linkage->sublinkage[linkage->current]);
-#else
-	Sublinkage *sublinkage = &(linkage->sublinkage);
-#endif /* USE_FAT_LINKAGES */
-	unsigned int N_links = sublinkage->num_links;
-	Link **ppla = sublinkage->link;
-	String * string;
+	unsigned int N_links = linkage->num_links;
+	Link *ppla = linkage->link_array;
+	String * string;  
 	char * gr_string;
-	Dictionary dict = linkage->sent->dict;
 	unsigned int N_words_to_print;
 
 	char picture[MAX_HEIGHT][MAX_LINE];
@@ -426,55 +392,41 @@ linkage_print_diagram_ctxt(const Linkage linkage,
 
 	/* Do we want to print the left wall? */
 	N_wall_connectors = 0;
-	if (dict->left_wall_defined)
+	suppressor_used = false;
+	if (!display_walls)
 	{
-		suppressor_used = false;
-		if (!display_walls)
-		{
-			for (j=0; j<N_links; j++)
-			{
-				if (0 == ppla[j]->lw)
-				{
-					if (ppla[j]->rw == linkage->num_words-1) continue;
-					N_wall_connectors ++;
-					if (0 == strcmp(ppla[j]->lc->string, LEFT_WALL_SUPPRESS))
-					{
-						suppressor_used = true;
-					}
-				}
-			}
-		}
-		print_word_0 = (((!suppressor_used) && (N_wall_connectors != 0))
-						|| (N_wall_connectors > 1) || display_walls);
-	}
-	else
-	{
-		print_word_0 = true;
-	}
-
-	/* Do we want to print the right wall? */
-	N_wall_connectors = 0;
-	if (dict->right_wall_defined)
-	{
-		suppressor_used = false;
 		for (j=0; j<N_links; j++)
 		{
-			if (ppla[j]->rw == linkage->num_words-1)
+			if (0 == ppla[j].lw)
 			{
+				if (ppla[j].rw == linkage->num_words-1) continue;
 				N_wall_connectors ++;
-				if (0 == strcmp(ppla[j]->lc->string, RIGHT_WALL_SUPPRESS))
+				if (0 == strcmp(ppla[j].lc->string, LEFT_WALL_SUPPRESS))
 				{
 					suppressor_used = true;
 				}
 			}
 		}
-		print_word_N = (((!suppressor_used) && (N_wall_connectors != 0))
-						|| (N_wall_connectors > 1) || display_walls);
 	}
-	else
+	print_word_0 = (((!suppressor_used) && (N_wall_connectors != 0))
+					|| (N_wall_connectors != 1) || display_walls);
+
+	/* Do we want to print the right wall? */
+	N_wall_connectors = 0;
+	suppressor_used = false;
+	for (j=0; j<N_links; j++)
 	{
-		print_word_N = true;
+		if (ppla[j].rw == linkage->num_words-1)
+		{
+			N_wall_connectors ++;
+			if (0 == strcmp(ppla[j].lc->string, RIGHT_WALL_SUPPRESS))
+			{
+				suppressor_used = true;
+			}
+		}
 	}
+	print_word_N = (((!suppressor_used) && (N_wall_connectors != 0))
+					|| (N_wall_connectors != 1) || display_walls);
 
 	N_words_to_print = linkage->num_words;
 	if (!print_word_N) N_words_to_print--;
@@ -492,16 +444,17 @@ linkage_print_diagram_ctxt(const Linkage linkage,
 	{
 		for (j=0; j<N_links; j++)
 		{
-			assert (ppla[j]->lw != SIZE_MAX);
-			if (((unsigned int) (ppla[j]->rw - ppla[j]->lw)) != link_length)
+			assert (ppla[j].lw != SIZE_MAX);
+			if (NULL == ppla[j].link_name) continue;
+			if (((unsigned int) (ppla[j].rw - ppla[j].lw)) != link_length)
 			  continue;
-			if (!print_word_0 && (ppla[j]->lw == 0)) continue;
+			if (!print_word_0 && (ppla[j].lw == 0)) continue;
 			/* Gets rid of the irrelevant link to the left wall */
-			if (!print_word_N && (ppla[j]->rw == linkage->num_words-1)) continue;
+			if (!print_word_N && (ppla[j].rw == linkage->num_words-1)) continue;
 
 			/* Put it into the lowest position */
-			cl = center[ppla[j]->lw];
-			cr = center[ppla[j]->rw];
+			cl = center[ppla[j].lw];
+			cr = center[ppla[j].rw];
 			for (row=0; row < MAX_HEIGHT; row++)
 			{
 				for (k=cl+1; k<cr; k++)
@@ -527,33 +480,31 @@ linkage_print_diagram_ctxt(const Linkage linkage,
 			for (k=cl+1; k<cr; k++) {
 				picture[row][k] = '-';
 			}
-			s = ppla[j]->link_name;
-
-#ifdef USE_FAT_LINKAGES
-			if (!is_utf8_alpha(s))
-			  s = "";  /* Don't print fat link connector name */
-#endif /* USE_FAT_LINKAGES */
 
+			s = ppla[j].link_name;
 			k = strlen(s);
-			if ((cl+cr-k)/2 + 1 <= cl) {
+			inc = cl + cr + 2;
+			if (inc < k) inc = 0;
+			else inc = (inc-k)/2;
+			if (inc <= cl) {
 				t = picture[row] + cl + 1;
 			} else {
-				t = picture[row] + (cl + cr + 2 - k)/2;
+				t = picture[row] + inc;
 			}
 
 			/* Add direction indicator */
 			// if (DEPT_CHR == ppla[j]->lc->string[0]) { *(t-1) = '<'; }
-			if (DEPT_CHR == ppla[j]->lc->string[0] &&
+			if (DEPT_CHR == ppla[j].lc->string[0] &&
 			    (t > &picture[row][cl])) { picture[row][cl+1] = '<'; }
-			if (HEAD_CHR == ppla[j]->lc->string[0]) { *(t-1) = '>'; }
+			if (HEAD_CHR == ppla[j].lc->string[0]) { *(t-1) = '>'; }
 
 			/* Copy connector name; stop short if no room */
 			while ((*s != '\0') && (*t == '-')) *t++ = *s++;
 
 			/* Add direction indicator */
 			// if (DEPT_CHR == ppla[j]->rc->string[0]) { *t = '>'; }
-			if (DEPT_CHR == ppla[j]->rc->string[0]) { picture[row][cr-1] = '>'; }
-			if (HEAD_CHR == ppla[j]->rc->string[0]) { *t = '<'; }
+			if (DEPT_CHR == ppla[j].rc->string[0]) { picture[row][cr-1] = '>'; }
+			if (HEAD_CHR == ppla[j].rc->string[0]) { *t = '<'; }
 
 			/* The direction indicators maye have clobbered these. */
 			picture[row][cl] = '+';
@@ -576,8 +527,8 @@ linkage_print_diagram_ctxt(const Linkage linkage,
 	if (print_word_0) k = 0; else k = 1;
 	for (; k<N_words_to_print; k++) {
 		s = linkage->word[k];
-		i=0;
-		while(*s != '\0') {
+		i = 0;
+		while (*s != '\0') {
 			*t++ = *s++;
 			i++;
 		}
@@ -694,6 +645,9 @@ char * linkage_print_diagram(const Linkage linkage, bool display_walls, size_t s
 {
 	ps_ctxt_t ctx;
 	if (!linkage) return NULL;
+
+	ctx.link_heights = (int *) alloca(linkage->num_links * sizeof(int));
+	ctx.row_starts = (int *) alloca(linkage->num_words * sizeof(int));
 	return linkage_print_diagram_ctxt(linkage, display_walls, screen_width, &ctx);
 }
 
@@ -723,12 +677,14 @@ static const char * header(bool print_ps_header);
 
 char * linkage_print_postscript(const Linkage linkage, bool display_walls, bool print_ps_header)
 {
-	char * ps, * qs;
+	char * ps, * qs, * ascii;
 	int size;
 
 	/* call the ascii printer to initialize the row size stuff. */
 	ps_ctxt_t ctx;
-	char * ascii = linkage_print_diagram_ctxt(linkage, display_walls, 8000, &ctx);
+	ctx.link_heights = (int *) alloca(linkage->num_links * sizeof(int));
+	ctx.row_starts = (int *) alloca(linkage->num_words * sizeof(int));
+	ascii = linkage_print_diagram_ctxt(linkage, display_walls, 8000, &ctx);
 	linkage_free_diagram(ascii);
 
 	ps = build_linkage_postscript_string(linkage, display_walls, &ctx);
@@ -748,8 +704,8 @@ void linkage_free_postscript(char * s)
 
 char * linkage_print_pp_msgs(Linkage linkage)
 {
-	if (linkage && linkage->info && linkage->info->pp_violation_msg)
-		return strdup(linkage->info->pp_violation_msg);
+	if (linkage && linkage->lifo.pp_violation_msg)
+		return strdup(linkage->lifo.pp_violation_msg);
 	return strdup("");
 }
 
@@ -1301,6 +1257,9 @@ void print_sentence_word_alternatives(Sentence sent, bool debugprint,
 						}
 					}
 
+					if (0 == strcmp(wt, EMPTY_WORD_MARK))
+						wt = EMPTY_WORD_DISPLAY;
+
 					/* Restore SUBSCRIPT_DOT for printing */
 					st = strrchr(wt, SUBSCRIPT_MARK);
 					if (st)
@@ -1311,17 +1270,21 @@ void print_sentence_word_alternatives(Sentence sent, bool debugprint,
 					}
 					if (debugprint) lgdebug(0, " %s", wt);
 
-					/* For now each word component is called "Token".
-					 * TODO: Its type can be decoded and a more precise
-					 * term (stem, prefix, etc.) can be used.
-					 * Display the features of the token */
-					if (NULL == tokenpos && (NULL != display))
+					/* Don't try to give info on the empty word. */
+					if (0 != strcmp(wt, EMPTY_WORD_DISPLAY))
 					{
-						printf("Token \"%s\" ", wt);
-						display(sent->dict, wt);
-						printf("\n");
+						/* For now each word component is called "Token".
+						 * TODO: Its type can be decoded and a more precise
+						 * term (stem, prefix, etc.) can be used.
+						 * Display the features of the token */
+						if (NULL == tokenpos && (NULL != display))
+						{
+							printf("Token \"%s\" ", wt);
+							display(sent->dict, wt);
+							printf("\n");
+						}
+						else if (word_split) printf("  %s", wt);
 					}
-					else if (word_split) printf("  %s", wt);
 				}
 			}
 			if (!alt_exists)
@@ -1334,3 +1297,43 @@ void print_sentence_word_alternatives(Sentence sent, bool debugprint,
 	if (debugprint) lgdebug(0, "\n");
 	else if (word_split) printf("\n");
 }
+
+/**
+ * Print a word, converting SUBSCRIPT_MARK to SUBSCRIPT_DOT.
+ */
+void print_with_subscript_dot(const char *s)
+{
+	const char *mark = strchr(s, SUBSCRIPT_MARK);
+	size_t len = NULL != mark ? (size_t)(mark - s) : strlen(s);
+
+	printf("%.*s%s%s ", (int)len,
+			  s, NULL != mark ? "." : "", NULL != mark ? mark+1 : "");
+}
+
+/**
+ *  Print the chosen_disjuncts words.
+ *  This is used for debug, e.g. for traking them in the Wordgraph display.
+ */
+void print_chosen_disjuncts_words(const Linkage lkg)
+{
+	size_t i;
+
+	lgdebug(+0, "Linkage %p (%zu words): ", lkg, lkg->num_words);
+	for (i = 0; i < lkg->num_words; i++)
+	{
+		Disjunct *cdj = lkg->chosen_disjuncts[i];
+		const char *djw; /* disjunct word - the chosen word */
+
+		if (NULL == cdj)
+			djw = "[]"; /* null word */
+		else if (0 == strcmp(cdj->string, EMPTY_WORD_MARK))
+			djw = EMPTY_WORD_DISPLAY;
+		else if ('\0' == cdj->string[0])
+			djw = "\\0"; /* null string - something is wrong */
+		else
+			djw = cdj->string;
+
+		print_with_subscript_dot(djw);
+	}
+	lgdebug(0, "\n");
+}
