@@ -184,6 +184,7 @@ static bool is_number(const char * s)
 }
 #endif
 
+#if defined HAVE_HUNSPELL || defined HAVE_ASPELL
 /**
  * Returns true if the word contains digits.
  */
@@ -203,6 +204,7 @@ static bool contains_digits(const char * s)
 	}
 	return false;
 }
+#endif /* defined HAVE_HUNSPELL || defined HAVE_ASPELL */
 
 /**
  * Accumulate different word-stemming possibilities.
@@ -236,8 +238,7 @@ void add_alternative(Sentence sent,
 	int numlist[] = { prefnum, stemnum, suffnum };
 	enum affixtype { PREFIX, STEM, SUFFIX, END };
 	enum affixtype at;
-	Dictionary afdict = sent->dict->affix_table; /* for INFIX_MARK only */
-	char infix_mark = INFIX_MARK;
+	char infix_mark = INFIX_MARK(sent->dict->affix_table);
 	char buff[MAX_WORD+2];
 
 	lgdebug(+3, "Word%s:",
@@ -349,6 +350,7 @@ void add_alternative(Sentence sent,
 	sent->t_count = t_count;
 }
 
+#if defined HAVE_HUNSPELL || defined HAVE_ASPELL
 /**
  * Return true if an alternative has been issued for the current word.
  * t_count==0 if and only if no alternative has been issued yet.
@@ -356,6 +358,7 @@ void add_alternative(Sentence sent,
 static bool word_has_alternative(Sentence sent) {
 	return (sent->t_count > 0);
 }
+#endif /* defined HAVE_HUNSPELL || defined HAVE_ASPELL */
 
 /**
  * Terminate issuing alternatives to the current input word.
@@ -461,12 +464,12 @@ static bool add_alternative_with_subscr(Sentence sent, const char * prefix,
                                         const char * word, const char * suffix)
 {
 	Dictionary dict = sent->dict;
-	Dictionary afdict = dict->affix_table; /* Known to be non-NULL. */
 	Afdict_class * stemsubscr_list =
-	   AFCLASS(afdict, AFDICT_STEMSUBSCR);
+	   AFCLASS(dict->affix_table, AFDICT_STEMSUBSCR);
 	const char ** stemsubscr = stemsubscr_list->string;
 	size_t stemsubscr_count = stemsubscr_list->length;
 	bool word_is_in_dict = false;
+	char infix_mark = INFIX_MARK(dict->affix_table);
 
 	if (0 == stemsubscr_count)
 	{
@@ -476,29 +479,31 @@ static bool add_alternative_with_subscr(Sentence sent, const char * prefix,
 		/* If this is not a morpheme split (INFIX_MARK==NUL), the word is not
 		 * considered to be in the dict. This is important since then it can
 		 * match a regex. For example: 1960's, which may get split to: 1960 's */
-		if ('\0' != INFIX_MARK) word_is_in_dict = true;
+		if ('\0' != infix_mark) word_is_in_dict = true;
 	}
 	else
 	{
 		size_t si;
-		size_t wlen = MIN(strlen(word), MAX_WORD);
-		char w[MAX_WORD*2];
-		const char * nwp = w;
-
-		strncpy(w, word, wlen);
-		w[wlen] = '\0';
+		size_t wlen = strlen(word);
+		size_t slen = 0;
+		char *w;
 
 		for (si = 0; si < stemsubscr_count; si++)
 		{
-			size_t slen = MIN(strlen(stemsubscr[si]), MAX_WORD-1);
+			slen = MAX(slen, strlen(stemsubscr[si]));
+		}
+		w = alloca(wlen + slen + 1);
+		strcpy(w, word);
 
-			strncpy(&w[wlen], stemsubscr[si], slen);
-			w[wlen+slen] = '\0';
+		for (si = 0; si < stemsubscr_count; si++)
+		{
+			strcpy(&w[wlen], stemsubscr[si]);
 
 			/* We should not match regexes to stems. */
 			if (boolean_dictionary_lookup(dict, w))
 			{
-				add_alternative(sent, (prefix ? 1 : 0),&prefix, 1,&nwp, 1,&suffix);
+				add_alternative(sent, (prefix ? 1 : 0),&prefix,
+				                1,(const char **)&w, 1,&suffix);
 				word_is_in_dict = true;
 			}
 		}
@@ -515,18 +520,18 @@ static bool add_alternative_with_subscr(Sentence sent, const char * prefix,
  * Note: If a word can split it doesn't say it is a real dictionary word,
  * as there can still be no links between some of its parts.
  *
- * XXX The prefix code is not validated yet by actual use.
+ * The prefix code is only lightly validated by actual use.
  */
 static bool suffix_split(Sentence sent, const char *w, const char *wend)
 {
-	int i, j, len;
+	int i, j;
 	Afdict_class *prefix_list, *suffix_list;
 	int p_strippable, s_strippable;
 	const char **prefix, **suffix;
 	const char *no_suffix = NULL;
-	char newword[MAX_WORD+1];
 	bool word_can_split = false;
 	Dictionary dict = sent->dict;
+	char *newword = alloca(wend-w+1);
 
 	/* Set up affix tables. */
 	if (NULL == dict->affix_table) return false;
@@ -544,18 +549,21 @@ static bool suffix_split(Sentence sent, const char *w, const char *wend)
 	 * prefixes only, without suffixes). */
 	for (i = 0; i <= s_strippable; i++, suffix++)
 	{
+		bool did_split = false;
+		size_t suflen = 0;
 		if (i < s_strippable)
 		{
-			len = strlen(*suffix);
-			/* The remaining w is too short for a possible match. */
-			if ((wend-w) < len) continue;
+			suflen = strlen(*suffix);
+			 /* The remaining w is too short for a possible match.
+			  * In addition, don't allow empty stems. */
+			if ((wend-suflen) < (w+1)) continue;
 
 			/* A lang like Russian allows empty suffixes, which have a real
 			 * morphological linkage. In the following check, the empty suffix
 			 * always matches. */
-			if (0 == strncmp(wend-len, *suffix, len))
+			if (0 == strncmp(wend-suflen, *suffix, suflen))
 			{
-				size_t sz = MIN((wend-len)-w, MAX_WORD);
+				size_t sz = MIN((wend-w)-suflen, MAX_WORD);
 				strncpy(newword, w, sz);
 				newword[sz] = '\0';
 
@@ -565,33 +573,42 @@ static bool suffix_split(Sentence sent, const char *w, const char *wend)
 				 * not boolean_dictionary_lookup(). */
 				if (find_word_in_dict(dict, newword))
 				{
-					word_can_split |=
+					did_split =
 						add_alternative_with_subscr(sent, NULL, newword, *suffix);
+					word_can_split |= did_split;
 				}
 			}
 		}
 		else
 		{
-				len = 0;
-				suffix = &no_suffix;
+			suflen = 0;
+			suffix = &no_suffix;
 		}
 
 		/*
-		 * Try stripping off prefixes.
-		 * XXX Not validated yet by actual use.
+		 * Try stripping off prefixes. Avoid double-counting and
+		 * other trouble by doing this only if we split off a suffix,
+		 * or if there is no suffix.
 		 */
-		for (j = 0; j < p_strippable; j++)
+		if (did_split || 0==suflen)
 		{
-			if (strncmp(w, prefix[j], strlen(prefix[j])) == 0)
+			for (j = 0; j < p_strippable; j++)
 			{
-				int sz = MIN((wend - len) - (w + strlen(prefix[j])), MAX_WORD);
-				strncpy(newword, w+strlen(prefix[j]), sz);
-				newword[sz] = '\0';
-				/* ??? Do we need a regex match? */
-				if (boolean_dictionary_lookup(dict, newword))
+				size_t prelen = strlen(prefix[j]);
+				/* The remaining w is too short for a possible match. */
+				if ((wend-w) - suflen < prelen) continue;
+				if (strncmp(w, prefix[j], prelen) == 0)
 				{
-					word_can_split |=
-					 add_alternative_with_subscr(sent, prefix[j], newword, *suffix);
+					size_t sz = MIN((wend-w) - suflen - prelen, MAX_WORD);
+
+					strncpy(newword, w+prelen, sz);
+					newword[sz] = '\0';
+					/* ??? Do we need a regex match? */
+					if (boolean_dictionary_lookup(dict, newword))
+					{
+						word_can_split |=
+					 	add_alternative_with_subscr(sent, prefix[j], newword, *suffix);
+					}
 				}
 			}
 		}
@@ -615,6 +632,12 @@ static bool suffix_split(Sentence sent, const char *w, const char *wend)
  * Add all the alternatives.
  * The assumptions used prevent a large number of false splits.
  * They may be relaxed later.
+ * 
+ * XXX Because the grammatical rules of which prefixes are valid for the
+ * remaining word are not checked, non-existing words may get split. In such a
+ * case there is no opportunity for a regex or spell check of this unknown word.
+ * FIXME Before issuing an alternative, validate that the combination is
+ * supported by the dict.
  *
  * Note: This function currently does more than absolutely needed for LG,
  * in order to simplify the initial Hebrew dictionary.
@@ -627,7 +650,6 @@ static bool suffix_split(Sentence sent, const char *w, const char *wend)
  * - subwords in a prefix are unique ('ככ' is considered here as one "subword")
  * - input words with length <= 2 don't have a prefix
  * - each character uses 2 bytes
- * - less than 16 prefix types - 13 are actually defined (bool array limit)
  * - the input word contains only Hebrew characters
  * - the letter "ו" (vav) can only be the first prefix subword
  * - if the last prefix subword is not "ו" and the word (length>2) starts
@@ -650,8 +672,7 @@ static bool mprefix_split(Sentence sent, const char *word)
 	bool word_is_in_dict;
 	int split_prefix_i = 0;      /* split prefix index */
 	const char *split_prefix[HEB_PRENUM_MAX]; /* the whole prefix */
-	/* pseen is a simple prefix combination filter */
-	bool pseen[HEB_MPREFIX_MAX]; /* prefix "subword" seen */
+	bool *pseen;                 /* prefix "subword" seen (not allowed again) */
 	Dictionary dict = sent->dict;
 	int wordlen;
 	int wlen;
@@ -662,14 +683,14 @@ static bool mprefix_split(Sentence sent, const char *word)
 	mprefix_list = AFCLASS(dict->affix_table, AFDICT_MPRE);
 	mp_strippable = mprefix_list->length;
 	if (0 == mp_strippable) return false;
-	assert(mp_strippable <= HEB_MPREFIX_MAX, /* sanity check */
-		"mp_strippable>" STRINGIFY(HEB_MPREFIX_MAX));
 	/* The mprefix list is revered-sorted according to prefix length.
 	 * The code here depends on that. */
 	mprefix = mprefix_list->string;
 
-	/* zero-out pseen[], assuming zeroed-out bytes are interpreted as false */
-	memset(pseen, 0, sizeof(pseen));
+	pseen = alloca(mp_strippable * sizeof(*pseen));
+	/* Assuming zeroed-out bytes are interpreted as false. */
+	memset(pseen, 0, mp_strippable * sizeof(*pseen));
+
 	word_is_in_dict = false;
 	w = word;
 	wordlen = strlen(word);  /* guaranteed < MAX_WORD by separate_word() */
@@ -888,16 +909,8 @@ static bool guess_misspelled_word(Sentence sent, const char * word,
 }
 #endif /* HAVE_HUNSPELL */
 
-/* Strip off punctuation, etc. on the left-hand side. */
-/* XXX FIXME: this fails in certain cases: e.g.
- * "By the '50s, he was very prosperous."
- * where the leading quote is striped, and then "50s," cannot be
- * found in the dict.  Next, the comma is removed, and "50s" is still
- * not in the dict ... the trick was that the comma should be
- * right-stripped first, then the possible quotes.
- * More generally, the current implementation of the link-parser algo
- * does not support multiple alternative tokenizations; the viterbi
- * parser, under development, should be able to do better.
+/**
+ * Strip off punctuation, etc. on the left-hand side.
  */
 static const char * strip_left(Sentence sent, const char * w,
                                bool quote_found)
@@ -1003,7 +1016,7 @@ static const char * strip_left(Sentence sent, const char * w,
  *
  * p is a mark of the invocation position, for debugging.
  */
-extern const char const *afdict_classname[]; /* For debug message only */
+extern const char *const afdict_classname[]; /* For debug message only */
 static bool strip_right(Sentence sent,
 		                  const char *w,
                         const char **wend,
@@ -1016,7 +1029,7 @@ static bool strip_right(Sentence sent,
 	Dictionary dict = sent->dict;
 	Dictionary afdict = dict->affix_table;
 	const char * temp_wend = *wend;
-	char *word = alloca(temp_wend-w);
+	char *word = alloca(temp_wend-w+1);
 	size_t sz;
 	size_t i;
 	size_t nrs = 0;
@@ -1046,7 +1059,7 @@ static bool strip_right(Sentence sent,
 
 			if (strncmp(temp_wend-len, t, len) == 0)
 			{
-				lgdebug(2, "%d: right strip %s: w='%s' rword '%s'\n",
+				lgdebug(2, "%d: strip_right(%s): w='%s' rword '%s'\n",
 				        p, afdict_classname[classnum], temp_wend-len, t);
 				r_stripped[*n_r_stripped+nrs] = t;
 				nrs++;
@@ -1142,6 +1155,7 @@ static void separate_word(Sentence sent, Parse_Options opts,
 	int downcase_size = sz+MB_LEN_MAX+1; /* pessimistic max. size of dc buffer */
 	char *downcase = alloca(downcase_size);               /* downcasing buffer */
 	char *str = alloca(downcase_size+sizeof("[!]"));        /* tmp word buffer */
+	char *seen_word = alloca(downcase_size);    /* loop-prevention word buffer */
 
 	downcase[0] = '\0';
 
@@ -1206,6 +1220,7 @@ static void separate_word(Sentence sent, Parse_Options opts,
 	 * units, then we try punctuation, and then units. This allows commas to be
 	 * removed (e.g.  7grams,). */
 
+	seen_word[0] = '\0';
 	do
 	{
 		int temp_n_r_stripped;
@@ -1257,13 +1272,17 @@ static void separate_word(Sentence sent, Parse_Options opts,
 		strncpy(word, w, sz);
 		word[sz] = '\0';
 
+		/* Avoid an infinite loop due to a repeating unknown remaining word */
+		if (0 == strcmp(word, seen_word)) break;
+		strcpy(seen_word, word);
+
 	/* Any remaining dict word stops the right-punctuation stripping. */
 	} while (NULL == units_wend && stripped &&
 	         !boolean_dictionary_lookup(dict, word));
 
 	/* Check whether the <number><units> "word" is in the dict
 	 * (including regex). In such a case we need to generate an alternative. */
-	if (units_wend) /* units found */
+	if (units_n_r_stripped && units_wend) /* units found */
 	{
 		sz = units_wend-w;
 		strncpy(str, w, sz);
@@ -1278,7 +1297,7 @@ static void separate_word(Sentence sent, Parse_Options opts,
 		for (i = n_r_stripped - 1; i >= 0; i--)
 			printf("[%d]='%s'%s", i, r_stripped[i], i>0 ? "," : "");
 	}
-	lgdebug(SWLEV, "), w='%s' wend='%s' word='%s' units_wend='%s' input_word='%s' "
+	lgdebug(0 SWLEV, "), w='%s' wend='%s' word='%s' units_wend='%s' input_word='%s' "
            "units_alternative=%d\n", w, wend, word, units_wend, input_word,
 			  units_alternative);
 
@@ -1300,7 +1319,7 @@ static void separate_word(Sentence sent, Parse_Options opts,
 		/* The input word is in the dict and also got stripped off
 		 * something - "word" here is what remained. */
 		word_is_in_dict = find_word_in_dict(dict, word);
-		lgdebug(SWLEV, "Check stripped word='%s' boolean_dictionary_lookup=%d "
+		lgdebug(SWLEV, "Check stripped word='%s' find_word_in_dict=%d "
 		        "is_utf8_digit=%d\n", word, word_is_in_dict, start_digit);
 		if (word_is_in_dict || units_alternative)
 		{
@@ -1514,7 +1533,7 @@ static void separate_word(Sentence sent, Parse_Options opts,
 		issued = guess_misspelled_word(sent, word, quote_found, opts);
 		lgdebug(SWLEV, "Spell suggest=%d\n", issued);
 	}
-#endif /* HAVE_HUNSPELL */
+#endif /* defined HAVE_HUNSPELL || defined HAVE_ASPELL */
 
 	if (false == issued)
 	{
