@@ -42,16 +42,18 @@
 #include <sys/ioctl.h>
 #include <fcntl.h>
 #include <unistd.h>
+#else
+#include <windows.h>
+#include <wchar.h>
 #endif
 
 #ifdef _MSC_VER
 #define LINK_GRAMMAR_DLL_EXPORT 0
 #endif
 
-#include <link-grammar/link-includes.h>
+#include "../link-grammar/link-includes.h"
 #include "command-line.h"
 #include "lg_readline.h"
-#include "../link-grammar/expand.h"
 #include "../link-grammar/utilities.h"     /* For MSVC portability */
 #include "../viterbi/viterbi.h"
 #include "../link-grammar/error.h"
@@ -245,70 +247,45 @@ static void process_linkage(Linkage linkage, Command_Options* copts)
 {
 	char * string;
 	ConstituentDisplayStyle mode;
-#ifdef USE_FAT_LINKAGES
-	Parse_Options opts = copts->popts;
-	int j, first_sublinkage;
-	int nlink;
-#endif /* USE_FAT_LINKAGES */
 
 	if (!linkage) return;  /* Can happen in timeout mode */
 
-#ifdef USE_FAT_LINKAGES
-	if (parse_options_get_use_fat_links(opts) &&
-	    copts->display_union)
+	if (copts->display_bad)
 	{
-		linkage_compute_union(linkage);
-		first_sublinkage = linkage_get_num_sublinkages(linkage)-1;
+		string = linkage_print_pp_msgs(linkage);
+		fprintf(stdout, "%s\n", string);
+		linkage_free_pp_msgs(string);
 	}
-	else
+	if (copts->display_on)
 	{
-		first_sublinkage = 0;
+		string = linkage_print_diagram(linkage, copts->display_walls, copts->screen_width);
+		fprintf(stdout, "%s", string);
+		linkage_free_diagram(string);
 	}
-
-	nlink = linkage_get_num_sublinkages(linkage);
-	for (j=first_sublinkage; j<nlink; ++j)
+	if (copts->display_links)
 	{
-		linkage_set_current_sublinkage(linkage, j);
-#endif /* USE_FAT_LINKAGES */
-		if (copts->display_bad)
-		{
-			string = linkage_print_pp_msgs(linkage);
-			fprintf(stdout, "%s\n", string);
-			linkage_free_pp_msgs(string);
-		}
-		if (copts->display_on)
-		{
-			string = linkage_print_diagram(linkage, copts->display_walls, copts->screen_width);
-			fprintf(stdout, "%s", string);
-			linkage_free_diagram(string);
-		}
-		if (copts->display_links)
-		{
-			string = linkage_print_links_and_domains(linkage);
-			fprintf(stdout, "%s", string);
-			linkage_free_links_and_domains(string);
-		}
-		if (copts->display_senses)
-		{
-			string = linkage_print_senses(linkage);
-			fprintf(stdout, "%s", string);
-			linkage_free_senses(string);
-		}
-		if (copts->display_disjuncts)
-		{
-			string = linkage_print_disjuncts(linkage);
-			fprintf(stdout, "%s", string);
-			linkage_free_disjuncts(string);
-		}
-		if (copts->display_postscript)
-		{
-			string = linkage_print_postscript(linkage, copts->display_walls, false);
-			fprintf(stdout, "%s\n", string);
-			linkage_free_postscript(string);
-		}
-#ifdef USE_FAT_LINKAGES
+		string = linkage_print_links_and_domains(linkage);
+		fprintf(stdout, "%s", string);
+		linkage_free_links_and_domains(string);
+	}
+	if (copts->display_senses)
+	{
+		string = linkage_print_senses(linkage);
+		fprintf(stdout, "%s", string);
+		linkage_free_senses(string);
+	}
+	if (copts->display_disjuncts)
+	{
+		string = linkage_print_disjuncts(linkage);
+		fprintf(stdout, "%s", string);
+		linkage_free_disjuncts(string);
+	}
+	if (copts->display_postscript)
+	{
+		string = linkage_print_postscript(linkage, copts->display_walls, false);
+		fprintf(stdout, "%s\n", string);
+		linkage_free_postscript(string);
 	}
-#endif /* USE_FAT_LINKAGES */
 	if ((mode = copts->display_constituents))
 	{
 		string = linkage_print_constituent_tree(linkage, mode);
@@ -405,43 +382,7 @@ static int process_some_linkages(Sentence sent, Command_Options* copts)
 				fprintf(stdout, "	Linkage %d, ", num_displayed+1);
 			}
 
-#ifdef USE_FAT_LINKAGES
-			if (!linkage_is_canonical(linkage))
-			{
-				fprintf(stdout, "non-canonical, ");
-			}
-			if (linkage_is_improper(linkage))
-			{
-				fprintf(stdout, "improper fat linkage, ");
-			}
-			if (linkage_has_inconsistent_domains(linkage))
-			{
-				fprintf(stdout, "inconsistent domains, ");
-			}
-#endif /* USE_FAT_LINKAGES */
-
 			corpus_cost = linkage_corpus_cost(linkage);
-#ifdef USE_FAT_LINKAGES
-			if (corpus_cost < 0.0f)
-			{
-				fprintf(stdout, "cost vector = (UNUSED=%d DIS=%5.2f FAT=%d AND=%d LEN=%d)\n",
-				       linkage_unused_word_cost(linkage),
-				       linkage_disjunct_cost(linkage),
-				       linkage_is_fat(linkage),
-				       linkage_and_cost(linkage),
-				       linkage_link_cost(linkage));
-			}
-			else
-			{
-				fprintf(stdout, "cost vector = (CORP=%6.4f UNUSED=%d DIS=%5.2f FAT=%d AND=%d LEN=%d)\n",
-				       corpus_cost,
-				       linkage_unused_word_cost(linkage),
-				       linkage_disjunct_cost(linkage),
-				       linkage_is_fat(linkage),
-				       linkage_and_cost(linkage),
-				       linkage_link_cost(linkage));
-			}
-#else
 			if (corpus_cost < 0.0f)
 			{
 				fprintf(stdout, "cost vector = (UNUSED=%d DIS=%5.2f LEN=%d)\n",
@@ -457,7 +398,6 @@ static int process_some_linkages(Sentence sent, Command_Options* copts)
 				       linkage_disjunct_cost(linkage),
 				       linkage_link_cost(linkage));
 			}
-#endif /* USE_FAT_LINKAGES */
 		}
 
 		process_linkage(linkage, copts);
@@ -588,9 +528,6 @@ static void setup_panic_parse_options(Parse_Options opts)
 	parse_options_set_min_null_count(opts, 1);
 	parse_options_set_max_null_count(opts, 100);
 	parse_options_set_max_parse_time(opts, 60);
-#ifdef USE_FAT_LINKAGES
-	parse_options_set_use_fat_links(opts, 0);
-#endif /* USE_FAT_LINKAGES */
 	parse_options_set_islands_ok(opts, 1);
 	parse_options_set_short_length(opts, 6);
 	parse_options_set_all_short_connectors(opts, 1);
@@ -619,20 +556,45 @@ static void print_usage(char *str)
  */
 static void check_winsize(Command_Options* copts)
 {
-/* Neither windows nor MSYS have the ioctl support needed for this. */
 #ifdef _WIN32
-	/* unsupported for now */
+	/* untested code .. does this actually work ??? */
+	HANDLE console;
+	CONSOLE_SCREEN_BUFFER_INFO info;
+
+	int fd = fileno(stdout);
+	if (!isatty(fd)) return;
+
+	/* Create a handle to the console screen. */
+	console = CreateFileW(L"CONOUT$", GENERIC_READ | GENERIC_WRITE,
+		FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING,
+		0, NULL);
+	if (console == INVALID_HANDLE_VALUE) goto fail;
+
+	/* Calculate the size of the console window. */
+	if (GetConsoleScreenBufferInfo(console, &info) == 0) goto fail;
+	CloseHandle(console);
+
+	copts->screen_width = info.srWindow.Right - info.srWindow.Left;
+	return;
+
+fail:
+	copts->screen_width = 79;
+	return;
 #else
 	struct winsize ws;
-	int fd = open("/dev/tty", O_RDWR);
+	int fd = fileno(stdout);
+
+	/* If ther is no controlling terminal, the fileno will fail. This
+	 * seems to happen while building docker images, I don't know why.
+	 */
+	if (fd < 0) return;
 
 	if (0 != ioctl(fd, TIOCGWINSZ, &ws))
 	{
-		perror("ioctl(/dev/tty, TIOCGWINSZ)");
-		close(fd);
+		if (isatty(fd))
+			perror("stdout: ioctl TIOCGWINSZ");
 		return;
 	}
-	close(fd);
 
 	/* printf("rows %i\n", ws.ws_row); */
 	/* printf("cols %i\n", ws.ws_col); */
@@ -761,9 +723,6 @@ int main(int argc, char * argv[])
 	}
 
 	check_winsize(copts);
-#ifdef _WIN32
-	parse_options_set_screen_width(opts, 79);
-#endif
 
 #if !defined(_MSC_VER) && !defined(__MINGW32__)
 	prt_error("Info: Using locale %s.", locale);
@@ -822,8 +781,9 @@ int main(int argc, char * argv[])
 		if (!copts->batch_mode) batch_in_progress = false;
 		if ('\0' != test[0])
 		{
-			/* In batch mode warn only once */
-			if (!batch_in_progress)
+			/* In batch mode warn only once.
+			 * In auto-next-linkage mode don't warn at all. */
+			if (!batch_in_progress && (NULL == strstr(test, ",auto-next-linkage,")))
 			{
 				fflush(stdout);
 				/* Remind the developer this is a test mode. */
